/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
public class Arc : Seed {
	Point centre;
	/**
	 * Parametric angles (not polar angles)
	 */
	double angle0;
	double angle1;
	double radius_x;
	double radius_y;
	double rotation;

	/**
	 * Memoize sincos rotation
	 */
	double sin_rotation;
	double cos_rotation;

	protected override void init_default_sub (Point p) {
		centre = p;
		radius_x = 0;
		radius_y = 0;
		angle0 = 0;
		angle1 = Math.PI * 2;
		set_rotation (0);
	}

	protected override void init_full_sub (double[] args) {
		assert (args.length >= 7);
		centre = Point () {x = args[0], y = args[1]};
		radius_x = args[2];
		radius_y = args[3];
		angle0 = args[4];
		angle1 = args[5];
		set_rotation (args[6]);
	}

	protected override string serialize_sub () {
		return "Arc (%s, %s) (%s, %s) (%s, %s) %s".printf (
			Util.format (centre.x, 4), Util.format (centre.y, 4),
			Util.format (radius_x, 4), Util.format (radius_y, 4),
			Util.format (angle0, 4), Util.format (angle1, 4),
			Util.format (rotation, 4));
	}

	void set_rotation (double rot) {
		rotation = rot;
		Math.sincos (rotation, out sin_rotation, out cos_rotation);
	}

	/**
	 * For an approximately circular arc, don't use distance_to_param. Use raw t directly.
	 */
	protected override void init_distance_to_parametric () {
		if (radius_x == 0 && radius_y == 0) return;
		//The following is commented out for the benefit of Seed.self_check_intersect
		//if (Math.fabs (radius_x - radius_y) / double.max (radius_x, radius_y) <= 0.0001) return;
		base.init_distance_to_parametric ();
	}

	protected override Point point_raw (double t) {
		double sin;
		double cos;
		Math.sincos ((1.0 - t) * angle0  +  t * angle1, out sin, out cos);
		// Point () {x = radius_x * cos, y = radius_y * sin} .rotate (rotation) .move (centre);
		double x = radius_x * cos;
		double y = radius_y * sin;
		return Point () {
			x = centre.x + x * cos_rotation - y * sin_rotation,
			y = centre.y + x * sin_rotation + y * cos_rotation};
	}

	public override Point distant_point (double x_distance, double y_distance) {
		// scale y_distance by the width at x_distance
		y_distance *= ((1 - x_distance) * width0  +  x_distance * width1);

		// angle at x_distance relative to the ellipse
		var t = parameter (x_distance);
		var angle = (1.0 - t) * angle0  +  t * angle1;
		double sin_angle;
		double cos_angle;
		Math.sincos (angle, out sin_angle, out cos_angle);

		// absolute angle, i.e. angle + rotation
		var sin = sin_angle * cos_rotation + cos_angle * sin_rotation;
		var cos = cos_angle * cos_rotation - sin_angle * sin_rotation;

		//  Point () {x = radius_x * cos_angle, y = radius_y * sin_angle}
		//               .rotate (rotation)
		//               .move (centre)
		//               .move by y_distance along the "perpendicular"
		var x = radius_x * cos_angle;
		var y = radius_y * sin_angle;
		return Point () {
			x = (x * cos_rotation - y * sin_rotation)  +  centre.x  +  y_distance * cos,
			y = (x * sin_rotation + y * cos_rotation)  +  centre.y  +  y_distance * sin};
	}

	public override void draw_direct (Cairo.Context cr, double scale) {
		var a0 = angle0;
		do {
			var a1 = double.min (Math.PI / 2, angle1 - a0);
			draw_small_arc (cr, scale, a0, a1);
			a0 += Math.PI / 2;
		} while (a0 < angle1);
	}

	/**
	 * Draw arc subtending an angle of less than 90 degrees.
	 *
	 * Reference: Richard DeVeneza's "How to determine the control points of a Bezier
	 * curve that approximates a small circular arc", Nov 2004
	 */
	void draw_small_arc (Cairo.Context cr, double scale, double a0, double sweep) {
		// Calculate points using DeVeneza's formula
		var p0 = Point () {x = Math.cos (sweep / 2), y = Math.sin (sweep / 2)};
		var p1 = Point () {x = (4 - p0.x) / 3, y = (1 - p0.x) * (3 - p0.x) / (3 * p0.y)};
		var p2 = p1.scale (1, -1);
		var p3 = p0.scale (1, -1);

		// Rotate by (a0 + sweep / 2) to put the arc segment in its rightful place,
		// then scale to make the shape right,
		// then rotate to orient the entire ellipse,
		// then translate to centre.
		var rot = a0 + sweep / 2;
		p0 = p0.rotate (rot) .scale (radius_x, radius_y) .rotate (rotation) .move (centre) .scale (scale, scale);
		p1 = p1.rotate (rot) .scale (radius_x, radius_y) .rotate (rotation) .move (centre) .scale (scale, scale);
		p2 = p2.rotate (rot) .scale (radius_x, radius_y) .rotate (rotation) .move (centre) .scale (scale, scale);
		p3 = p3.rotate (rot) .scale (radius_x, radius_y) .rotate (rotation) .move (centre) .scale (scale, scale);

		cr.move_to (p0.x, p0.y);
		cr.curve_to (p1.x, p1.y, p2.x, p2.y, p3.x, p3.y);
		cr.stroke ();
	}

	public override void draw_controls (Cairo.Context cr, double scale) {
		var corner0 = Point () {x = -radius_x, y = -radius_y} .rotate (rotation) .move (centre) .scale (scale, scale);
		var corner1 = Point () {x = +radius_x, y = -radius_y} .rotate (rotation) .move (centre) .scale (scale, scale);
		var corner2 = Point () {x = +radius_x, y = +radius_y} .rotate (rotation) .move (centre) .scale (scale, scale);
		var corner3 = Point () {x = -radius_x, y = +radius_y} .rotate (rotation) .move (centre) .scale (scale, scale);
		draw_line_control (cr, corner0, corner1);
		draw_line_control (cr, corner1, corner2);
		draw_line_control (cr, corner2, corner3);
		draw_line_control (cr, corner3, corner0);
		draw_square_control (cr, corner2);
		draw_circle_control (cr, corner1);
		draw_circle_control (cr, point (0) .scale (scale, scale));
		draw_circle_control (cr, point (1) .scale (scale, scale));
	}

	/**
	 * Control:
	 *  0 : radius_x and radius_y
	 *  1 : rotation
	 *  2 : angle0
	 *  3 : angle1
	 */
	Point centre_save;
	Point topleft_save;

	protected override int pressed_control (Point p, double scale) {
		var corner1 = Point () {x = +radius_x, y = -radius_y} .rotate (rotation) .move (centre);
		var corner2 = Point () {x = +radius_x, y = +radius_y} .rotate (rotation) .move (centre);
		return within_touch_range (p, corner2, scale) ? 0 :
			   within_touch_range (p, corner1, scale) ? 1 :
			   within_touch_range (p, point (0), scale) ? 2 :
			   within_touch_range (p, point (1), scale) ? 3 : -1;
	}

	protected override void save_drag_context () {
		centre_save = centre;
		topleft_save = Point () {x = -radius_x, y = -radius_y} .rotate (rotation) .move (centre);
	}

	protected override void drag_control (Point p, double scale, bool snap_on, double grid_size, double snap_angle) {
		// Convert to radian
		snap_angle = Math.PI * snap_angle / 180;
		if (snap_on && selected_control == 0) snap (ref p, scale, grid_size);
		var canonical_p = Point () {x = p.x - centre.x, y = p.y - centre.y} .rotate (-rotation);
		var rx = canonical_p.x;
		var ry = canonical_p.y;
		if (selected_control == 0) {
			var size = Point () {x = p.x - topleft_save.x, y = p.y - topleft_save.y}. rotate (-rotation);
			radius_x = double.max (0, size.x / 2);
			radius_y = double.max (0, size.y / 2);
			centre = Point () {x = radius_x, y = radius_y} .rotate (rotation) .move (topleft_save);
		}
		else if (selected_control == 1) {
			// We have to correct for the fact that our control is NOT at angle zero
			set_rotation (Math.atan2 (p.y - centre.y, p.x - centre.x) + Math.atan2 (radius_y, radius_x));
			if (snap_on) rotation = Math.round (rotation / snap_angle) * snap_angle;
		}
		else if (selected_control == 2) {
			if (snap_on && within_snap_range (p, point (1), scale)) {
				// Snap to full ellipse
				angle0 = angle1;
				angle1 += 2 * Math.PI;
			}
			else {
				// Polar to parametric angle: tan t = (rx / ry) * tan polar
				angle0 = Math.atan2 (radius_x * ry, radius_y * rx);
				if (snap_on) angle0 = Math.round (angle0 / snap_angle) * snap_angle;
				normalize_angles ();
			}
		}
		else if (selected_control == 3) {
			if (snap_on && within_snap_range (p, point (0), scale)) {
				// Snap to full ellipse
				angle1 = angle0 + 2 * Math.PI;
			}
			else {
				// Polar to parametric angle: tan t = (rx / ry) * tan polar
				angle1 = Math.atan2 (radius_x * ry, radius_y * rx);
				if (snap_on) angle1 = Math.round (angle1 / snap_angle) * snap_angle;
				normalize_angles ();
			}
		}
		else assert_not_reached ();
	}

	protected override void move_shape (double delta_x, double delta_y) {
		centre = centre_save.move_xy (delta_x, delta_y);
	}

	protected override Point[] points_to_snap () {
		return {
			centre,
			Point () {x = -radius_x, y = 0} .rotate (rotation) .move (centre),
			Point () {x = radius_x, y = 0} .rotate (rotation) .move (centre),
			Point () {x = 0, y = -radius_y} .rotate (rotation) .move (centre),
			Point () {x = 0, y = radius_y} .rotate (rotation) .move (centre) };
	}

	void normalize_angles () {
		while (angle0 < 0) angle0 += 2 * Math.PI;
		while (angle0 > 2 * Math.PI) angle0 -= 2 * Math.PI;
		while (angle1 < 0) angle1 += 2 * Math.PI;
		while (angle1 > 2 * Math.PI) angle1 -= 2 * Math.PI;
		// Cross over angle=0 line?
		if (angle0 > angle1) angle1 += 2 * Math.PI;
	}

	public override void draw_icon (Cairo.Context cr, double width, double height) {
		double[] args = {width / 2, height / 2, width / 2.3, height / 2.3, 0, 1.4 * Math.PI, 0};
		init_full (args);
		draw_direct (cr, 1);
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
