/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
public class Curve : Seed {
	Point p0;
	Point p1;
	Point p2;
	Point p3;

	protected override void init_default_sub (Point p) {
		var disp = Math.fabs (p.x) < 1.5 ? 1 : 500;  // Ugly hack to make the controls visible
		p0 = p;
		p1 = p.move_xy (0, -0.2 * disp);
		p2 = p.move_xy (0, -0.2 * disp);
		p3 = p;
	}

	protected override void init_full_sub (double[] args) {
		assert (args.length >= 8);
		p0 = Point () {x = args[0], y = args[1]};
		p1 = Point () {x = args[2], y = args[3]};
		p2 = Point () {x = args[4], y = args[5]};
		p3 = Point () {x = args[6], y = args[7]};
	}

	protected override string serialize_sub () {
		return "Curve (%s, %s) (%s, %s) (%s, %s) (%s, %s)".printf (
			Util.format (p0.x, 4), Util.format (p0.y, 4),
			Util.format (p1.x, 4), Util.format (p1.y, 4),
			Util.format (p2.x, 4), Util.format (p2.y, 4),
			Util.format (p3.x, 4), Util.format (p3.y, 4));
	}

	public override Point point_raw (double t) {
		// (1-t)^3 p0  +  3 (1-t)^2 t p1  +  3 (1-t) t^2 p2  +  t^3 p3
		var s = 1.0 - t;
		var a = s*s*s;
		var b = 3 * s*s * t;
		var c = 3 * s * t*t;
		var d = t*t*t;
		var x = a * p0.x + b * p1.x + c * p2.x + d * p3.x;
		var y = a * p0.y + b * p1.y + c * p2.y + d * p3.y;
		return Point () {x = x, y = y};
	}

	public override Point distant_point (double x_distance, double y_distance) {
		// scale y_distance by the width at x_distance
		y_distance *= ((1 - x_distance) * width0  +  x_distance * width1);
		var t = parameter (x_distance);
		var p = point_raw (t);
		// d/dt of (1-t)^3 p0  +  3 (1-t)^2 t p1  +  3 (1-t) t^2 p2  +  t^3 p3
		var s = 1.0 - t;
		var a = -3 * s*s;
		var b = -6 * s * t  +  3 * s*s;
		var c = -3 * t*t  +  6 * s * t;
		var d = 3 * t*t;
		// slope of the tangent
		var x = a * p0.x + b * p1.x + c * p2.x + d * p3.x;
		var y = a * p0.y + b * p1.y + c * p2.y + d * p3.y;
		// rotate by -90 degree to get the perpendicular
		double sin;
		double cos;
		Math.sincos (Math.atan2 (-x, y), out sin, out cos);
		// translate p by y_distance along the perpendicular
		p.x += y_distance * cos;
		p.y += y_distance * sin;
		return p;
	}

	public override void draw_direct (Cairo.Context cr, double scale) {
		Point c0 = p0 .scale (scale, scale);
		Point c1 = p1 .scale (scale, scale);
		Point c2 = p2 .scale (scale, scale);
		Point c3 = p3 .scale (scale, scale);
		cr.move_to (c0.x, c0.y);
		cr.curve_to (c1.x, c1.y, c2.x, c2.y, c3.x, c3.y);
		cr.stroke ();
	}

	public override void draw_controls (Cairo.Context cr, double scale) {
		draw_line_control (cr, p0 .scale (scale, scale), p1 .scale (scale, scale));
		draw_line_control (cr, p2 .scale (scale, scale), p3 .scale (scale, scale));
		draw_square_control (cr, p0 .scale (scale, scale));
		draw_square_control (cr, p1 .scale (scale, scale));
		draw_square_control (cr, p2 .scale (scale, scale));
		draw_square_control (cr, p3 .scale (scale, scale));
	}

	public override int pressed_control (Point p, double scale) {
		return within_touch_range (p, p2, scale) ? 2 :
			   within_touch_range (p, p1, scale) ? 1 :
			   within_touch_range (p, p3, scale) ? 3 :
			   within_touch_range (p, p0, scale) ? 0 : -1;
	}

	Point p0_save;
	Point p1_save;
	Point p2_save;
	Point p3_save;

	protected override void save_drag_context () {
		p0_save = p0;
		p1_save = p1;
		p2_save = p2;
		p3_save = p3;
	}

	protected override void drag_control (Point p, double scale, bool snap_on, double grid_size, double snap_angle) {
		var p_snapped = p;
		if (snap_on) snap (ref p_snapped, scale, grid_size);
		// When dragging an end point, snap to the other end point
		// if it is within range.
		if (selected_control == 0) {
			p0 = snap_on && within_snap_range (p, p3, scale) ? p3 : p_snapped;
			// Move p1 relative to p0
			p1 = p1_save.move_xy (p0.x - p0_save.x, p0.y - p0_save.y);
		}
		else if (selected_control == 1) {
			p1 = p_snapped;
		}
		else if (selected_control == 2) {
			p2 = p_snapped;
		}
		else if (selected_control == 3) {
			p3 = snap_on && within_snap_range (p, p0, scale) ? p0 : p_snapped;
			// Move p2 relative to p3
			p2 = p2_save.move_xy (p3.x - p3_save.x, p3.y - p3_save.y);
		}
		else assert_not_reached ();
	}

	protected override void move_shape (double delta_x, double delta_y) {
		p0 = p0_save.move_xy (delta_x, delta_y);
		p1 = p1_save.move_xy (delta_x, delta_y);
		p2 = p2_save.move_xy (delta_x, delta_y);
		p3 = p3_save.move_xy (delta_x, delta_y);
	}

	protected override Point[] points_to_snap () {
		return {p0, p1, p2, p3};
	}

	public override void draw_icon (Cairo.Context cr, double width, double height) {
		double[] args = {0, 0, width / 4, height, 0.75 * width, 0, width, height};
		init_full (args);
		draw_direct (cr, 1);
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
