/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * Queue Cairo instructions.
 * Purpose:
 *   - Allow sequential cairo shapes to be computed in parallel. Only the
 *     render method needs to be called in proper sequence.
 * Limitations:
 *   - Only one move_to and this must be called before line_to
 */
public class DeferredCairo {
	Point[] points = new Point[16];
	/**
	 * type:
	 *   'f' : fill
	 *   anything not 'f' : stroke (default)
	 */
	char type;

	Cairo.ImageSurface image_surface;
	double x1;
	double y1;
	double width;
	double height;

	public void recycle () {
		image_surface = null;
		if (points.length > 16) points = new Point[16];
		points.length = 0;
	}

	public void render (Cairo.Context cr) {
		if (image_surface != null) {
			cr.translate (x1, y1);
			cr.set_source_surface (image_surface, 0, 0);
			cr.paint ();
			cr.translate (-x1, -y1);
		}
		else {
			if (points.length == 0) return;
			cr.move_to (points[0].x, points[0].y);
			for (int i = 1; i < points.length; i++) cr.line_to (points[i].x, points[i].y);
			if (type == 'f') cr.fill_preserve ();
			cr.stroke ();
		}
	}

	public void pixelize (double line_width, RGBA color) {
		if (points.length <= 6) return;
		var x1 = points[0].x;
		var y1 = points[0].y;
		var x2 = x1;
		var y2 = y1;
		for (int i = 1; i < points.length; i++) {
			var x = points[i].x;
			var y = points[i].y;
			if (x < x1) x1 = x;
			else if (x > x2) x2 = x;
			if (y < y1) y1 = y;
			else if (y > y2) y2 = y;
		}
		this.x1 = x1 = Math.floor (x1) - 1;
		this.y1 = y1 = Math.floor (y1) - 1;
		width =  Math.ceil (x2) + 2 - x1;
		height = Math.ceil (y2) + 2 - y1;
		if (width * height > 500) return;
		var image_surface = new Cairo.ImageSurface (Cairo.Format.ARGB32, (int) width, (int) height);
		var cr = new Cairo.Context (image_surface);
		cr.translate (-x1, -y1);
		cr.set_line_width (line_width);
		cr.set_source_rgba (color.r, color.g, color.b, color.a);
		render (cr);
		this.image_surface = image_surface;
	}

	public void move_to (Point p) {
		//points = new Point[16];
		points[0] = p;
		points.length = 1;
	}

	public void line_to (Point p) {
		if (points.length >= 100000) return;
		points += p;
	}

	public void stroke () {
		type = 's';
	}

	public void fill () {
		type = 'f';
	}

}
// vim: tabstop=4: textwidth=100: autoindent:
