/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
public enum Flip {
	NONE = 0,
	Y = 1,
	X = 2,
	XY = 3;

	public bool x () {
		return (this & X) == X;
	}

	public bool y () {
		return (this & Y) == Y;
	}

	/**
	 * bit: 0 => y, 1 => x
	 */
	public bool is_set (int bit) {
		return (this & (1 << bit)) != 0;
	}

	public Flip @set (int bit) {
		return this | (1 << bit);
	}

	public Flip clear (int bit) {
		return this & ~(1 << bit);
	}
}

public struct Tweak {
	public double scale_x;
	public double scale_y;
	public double rotate;   // in degree
	public double move_x;
	public double move_y;

	public Tweak () {
		scale_x = 1;
		scale_y = 1;
		rotate = 0;
		move_x = 0;
		move_y = 0;
	}

	public bool equals (Tweak t) {
		var epsilon = 0.00001;
		return
			Math.fabs (scale_x - t.scale_x) < epsilon &&
			Math.fabs (scale_y - t.scale_y) < epsilon &&
			Math.fabs (rotate - t.rotate) < epsilon &&
			Math.fabs (move_x - t.move_x) < epsilon &&
			Math.fabs (move_y - t.move_y) < epsilon;
	}

	public string serialize () {
		return "%s %s,  %s,  %s %s".printf (
			Util.format (scale_x, 3), Util.format (scale_y, 3),
			Util.format (rotate, 1),
			Util.format (move_x, 3), Util.format (move_y, 3));
	}
}

public class TweakWidget : Gtk.Button {
	public signal void tweaked ();
	Gtk.Window parent_window;
	Flip[] flip;
	Tweak tweak;
	Gtk.DrawingArea icon;

	/**
	 * If the user cancel the operation, restore these values.
	 */
	Flip[] flip_save;
	Tweak tweak_save;

	/**
	 * If full_tweak is false, show only flip options
	 */
	bool full_tweak;

	/**
	 * If false, don't emit tweak signal
	 */
	bool ready;

	/**
	 * Dialog components
	 */
	Gtk.Dialog dialog = null;
	Gtk.DrawingArea segment_view;
	Gtk.SpinButton scale_x_entry;
	Gtk.SpinButton scale_y_entry;
	Gtk.SpinButton rotate_entry;
	Gtk.SpinButton move_x_entry;
	Gtk.SpinButton move_y_entry;

	/**
	 * dialog_flip[segment] where segment = 0..1
	 */
	Flip[] dialog_flip;
	Tweak dialog_tweak;

	/**
	 * toggle[segment][bit] where segment = 0..1 and bit = 0..2
	 */
	Gtk.ToggleButton[,] toggle;
	int num_bits = 2;   // Actually 3, but the D bit is superfluous

	public TweakWidget (Gtk.Window parent_window, bool full_tweak) {
		this.parent_window = parent_window;
		flip = {Flip.NONE, Flip.Y};
		this.full_tweak = full_tweak;
		tweak = Tweak ();
		image = (icon = new Gtk.DrawingArea ());
		var gap = 4;
		var size = 20;
		icon.set_size_request ((size + gap) * 2, size + gap * 2);
		icon.set_has_window (false);
		icon.draw.connect ((cr) => {
			draw_segment (cr, tweak, flip[0], gap, gap, size);
			draw_segment (cr, tweak, flip[1], gap + size, gap, size);
			return true;
		});
		this.tweaked.connect (() => { icon.queue_draw (); });
		this.clicked.connect (() => { show_dialog (); });
	}

	public void set_tweaks (Tweak t, Flip flip0, Flip flip1) {
		if (flip0 != flip[0] || flip1 != flip[1] || !tweak.equals (t)) {
			flip[0] = flip0;
			flip[1] = flip1;
			if (full_tweak) tweak = t;
			tweaked ();
		}
	}

	public void get_tweaks (ref Tweak t, out Flip flip0, out Flip flip1) {
		flip0 = flip[0];
		flip1 = flip[1];
		if (full_tweak) t = tweak;
	}

	void show_dialog () {
		ready = false;
		if (dialog == null) make_dialog ();
		// Show current tweaks
		if (full_tweak) {
			scale_x_entry.@value = tweak.scale_x;
			scale_y_entry.@value = tweak.scale_y;
			rotate_entry.@value = tweak.rotate;
			move_x_entry.@value = tweak.move_x;
			move_y_entry.@value = tweak.move_y;
		}
		// Show current flips
		for (int seg = 0; seg < 2; seg++) {
			dialog_flip[seg] = flip[seg];
			for (int bit = 0; bit < num_bits; bit++) {
				toggle[seg, bit].active = dialog_flip[seg].is_set (bit);
			}
		}
		// Save in case the tweak dialog is cancelled
		flip_save = {flip[0], flip[1]};
		tweak_save = tweak;
		ready = true;
		var result = dialog.run ();
		dialog.hide ();
		if (result != Gtk.ResponseType.OK) set_tweaks (tweak_save, flip_save[0], flip_save[1]);
	}

	void make_dialog () {
		var title = full_tweak ? _("Tweak") : _("Flip");
		dialog = Util.make_dialog (parent_window, title, _("_OK"), (grid) => {
			dialog_flip = new Flip[2];
			dialog_tweak = Tweak ();

			if (full_tweak) grid.attach (create_tweak_entries (), 0, 0, 3, 1);

			// View how an 'F' is transformed
			segment_view = new Gtk.DrawingArea ();
			grid.attach (segment_view, 1, 1, 1, num_bits);
			var gap = 10;
			var size = 50;
			segment_view.set_size_request ((size + gap) * 2, size + gap * 2);
			segment_view.set_has_window (false);
			segment_view.halign = Gtk.Align.CENTER;
			segment_view.valign = Gtk.Align.CENTER;
			segment_view.draw.connect ((cr) => {
				draw_segment (cr, dialog_tweak, dialog_flip[0], gap, gap, size);
				draw_segment (cr, dialog_tweak, dialog_flip[1], gap + size, gap, size);
				return true;
			});

			// Toggle buttons
			toggle = new Gtk.ToggleButton[2,num_bits];
			for (var seg = 0; seg < 2; seg++) {
				for (var bit = 0; bit < num_bits; bit++) {
					var t = create_button (seg, bit);
					t.halign = Gtk.Align.CENTER;
					t.valign = Gtk.Align.CENTER;
					toggle[seg, bit] = t;
					var col = seg * 2;
					var row = 1 + bit;
					grid.attach (t, col, row, 1, 1);
				}
			}
		});
	}

	/**
	 * Create entries for Scale, Rotate and Move.
	 */
	Gtk.Widget create_tweak_entries () {
		var grid = new Gtk.Grid ();
		grid.column_spacing = 5;
		grid.row_spacing = 5;
		grid.margin_bottom = 20;
		attach_entry (grid, 0, _("Scale:"), {scale_x_entry = Util.make_spin_button(0.25, 2.0, 0.005),
			                                 scale_y_entry = Util.make_spin_button(0.25, 2.0, 0.005)});
		attach_entry (grid, 1, _("Rotate:"), {rotate_entry = Util.make_spin_button(0, 360, 0.5)});
		attach_entry (grid, 2, _("Move:"), {move_x_entry = Util.make_spin_button(-1.0, 1.0, 0.005),
			                                move_y_entry = Util.make_spin_button(-1.0, 1.0, 0.005)});
		return grid;
	}

	/**
	 * Attach an entry. This is part of create_tweak_entries.
	 */
	void attach_entry (Gtk.Grid grid, int y, string? label_text, Gtk.SpinButton?[] buttons) {
		if (label_text != null) grid.attach (Util.label(label_text), 0, y, 1, 1);
		var x = 1;
		foreach (var button in buttons) {
			grid.attach (button, x++, y, 1, 1);
			button.value_changed.connect (() => {
				dialog_tweak.scale_x = scale_x_entry.@value;
				dialog_tweak.scale_y = scale_y_entry.@value;
				dialog_tweak.rotate = rotate_entry.@value;
				dialog_tweak.move_x = move_x_entry.@value;
				dialog_tweak.move_y = move_y_entry.@value;
				segment_view.queue_draw ();
				if (ready) {
					set_tweaks (dialog_tweak, dialog_flip[0], dialog_flip[1]);
				}
			});
		}
	}

	/**
	 * Create button for toggling flip.
	 * segment can be 0 or 1 (even or odd segments).
	 * Possible value for bit: see Flip.is_set
	 */
	Gtk.ToggleButton create_button (int segment, int bit) {
		var button = new Gtk.ToggleButton ();
		button.set_mode (false);
		button.border_width = 0;
		var image = new Gtk.DrawingArea ();
		var size = 15;
		image.set_size_request (size, size);
		image.set_has_window (false);
		image.draw.connect ((cr) => {
			cr.translate (0.5 * size, 0.5 * size);
			cr.scale (size, size);
			if (bit == 1) cr.rotate (Math.PI / 2);
			if (bit == 2) cr.rotate (Math.PI / 4);
			// Draw a 2-way arrow
			cr.move_to (0.0, -0.5);
			cr.line_to (0.3, -0.1);
			cr.line_to (0.1, -0.1);
			cr.line_to (0.1, 0.1);
			cr.line_to (0.3, 0.1);
			cr.line_to (0.0, 0.5);
			cr.line_to (-0.3, 0.1);
			cr.line_to (-0.1, 0.1);
			cr.line_to (-0.1, -0.1);
			cr.line_to (-0.3, -0.1);
			cr.line_to (0.0, -0.5);
			cr.fill ();
			return true;});
		button.image = image;
		// Initial state of the button
		button.active = flip[segment].is_set (bit);
		// Toggle the appropriate bit of the appropriate segment
		button.toggled.connect (() => {
			dialog_flip[segment] = (button.active ? dialog_flip[segment].@set (bit) :
			                                        dialog_flip[segment].clear (bit));
			segment_view.queue_draw ();
			if (ready) {
				set_tweaks (dialog_tweak, dialog_flip[0], dialog_flip[1]);
			}
		});
		return button;
	}

	/**
	 * Draw an F to indicate how the segment is tweaked
	 */
	void draw_segment (Cairo.Context cr, Tweak tweak, Flip flip, double x, double y, double size) {
		cr.save ();
		// Show segment boundary
		cr.rectangle (x, y, size, size);
		cr.set_line_width (0.2);
		cr.stroke ();

		// The centre of F becomes (0, 0) and the size is 1x1
		cr.translate (x + size / 2, y + size / 2);
		cr.scale (size, size);
		// Apply flip
		if (flip.x ()) cr.scale (-1, 1);
		if (flip.y ()) cr.scale (1, -1);
		// Apply other tweaks
		cr.translate (tweak.move_x, tweak.move_y);
		cr.rotate (tweak.rotate * Math.PI / 180);
		cr.scale (tweak.scale_x, tweak.scale_y);
		// Draw F
		cr.move_to (-0.4, -0.4);
		cr.line_to (0.4, -0.4);
		cr.line_to (0.4, -0.2);
		cr.line_to (-0.2, -0.2);
		cr.line_to (-0.2, -0.1);
		cr.line_to (0.2, -0.1);
		cr.line_to (0.2, 0.1);
		cr.line_to (-0.2, 0.1);
		cr.line_to (-0.2, 0.4);
		cr.line_to (-0.4, 0.4);
		cr.fill ();
		cr.restore ();
	}
}

// vim: tabstop=4: textwidth=100: autoindent:
