/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
using Gee;

public class Fractalizer : Object {
	/**
	 * A fractal is a tree. This class specifies a node in that tree.
	 * The tree is flattened into a linked list.
	 *
	 * Starting from a single root node, several threads cooperate to
	 * expand the tree. One of the threads holds a baton. The baton holder
	 * has the responsibility to make the final sequential rendition using
	 * cairo.
	 *
	 * All threads are equal! If a baton holder encounters a node that
	 * is already claimed by another thread and this other thread has not
	 * finished expanding the node, rather than waiting the baton holder
	 * must give up its baton and move on to expand another node.
	 */
	class Node : Projection {
		/**
		 * Ref count for custom memory management
		 *    count_left_link: the number of threads whose left_most points to this Node
		 *                     plus 1 for the previous Node's next.
		 *                     When this is 0, dec next->count_left_link.
		 *                     Also, dispose if this is a leaf.
		 *                     (See Fractalizer.advance_pointer)
		 *    last_child: If this is true, dispose the parent as well.
		 */
		public int count_left_link;
		public bool last_child;
		/**
		 * Not to be confused with Projection.parent which may be optimized away
		 * by Projection.compose.
		 */
		public Node* parent;
		
		bool leaf_computed;
		public bool leaf;
		DeferredCairo dc;
		/**
		 * A node is expanded if its childs have been generated,
		 * or if it turns out to be a leaf
		 */
		public bool expanded;

		weak Fractalizer fractalizer;
		public int depth;
		weak Seed seed;
		int level_of_detail;
		RGBA color;
		RGBA base_color_prex;        // Base color pre-multiplied by its mix
		double base_mix_cmp;        // 1 - mix value of base color
		bool twisted;
		Node* next;    // Next node in the fractal tree. Access is mutexed
		bool claimed;  // Accessed in lock(claimed)
		bool baton_passing;  // Accessed in lock(claimed)

		double width;   // calculated by Seed.too_thin

		/**
		 * Memory pool.
		 *   count_stack: used to limit the stack size
		 *   create: create new Node only if there is none left inside the pool
		 *   dispose: return Node to the pool
		 */
		Node* stack_next;
		static Node* stack_top = null;
		static int count_stack;

		public static Node* create () {
			Node* node;
			lock (stack_top) {
				node = stack_top;
				if (node == null) {
					node = new Node ();
					node->dc = new DeferredCairo ();
				}
				else {
					stack_top = stack_top->stack_next;
					count_stack--;
				}
			}
			node->count_left_link = 1;
			node->last_child = false;
			node->leaf_computed = false;
			node->expanded = false;
			node->claimed = false;
			node->baton_passing = false;
			return node;
		}

		public static Node* create_root (Fractalizer fractalizer, Seed seed) {
			Node* node = create ();
			node->init (
			      fractalizer, 0, null, null, seed, 
			      fractalizer.quality_rendering ? seed.level_of_detail + 1 : 1,
			      RGBA (0, 0, 0, 0), 1, RGBA (0, 0, 0, 0), false, null);
			return node;
		}
		
		public static void dispose (Node* node) {
			node->dc.recycle ();     // Reclaim memory used by DeferredCairo.points
			lock (stack_top) {
				if (count_stack < 10000) {
					node->stack_next = stack_top;
					stack_top = node;
					count_stack++;
				}
				else {
					delete node;
				}
			}
		}
		
		public void init (Fractalizer fractalizer, int depth, Projection* pr, Node* parent, Seed seed,
		             int level_of_detail, RGBA color, double base_mix_cmp, RGBA base_color_prex,
		             bool twisted, Node* next_arg) {
		    this.parent = parent;
			this.depth = depth;
			switch (depth) {
				case 0 :
					compose_unit_matrix ();
					break;
				case 1 :
					compose (null, pr);   // Optimize away the unit matrix at depth 0
					break;
				default :
					compose (parent, pr);
					break;
			}
			this.fractalizer = fractalizer;
			this.seed = seed;
			this.level_of_detail = level_of_detail;
			this.color = color;
			this.base_mix_cmp = base_mix_cmp;
			this.base_color_prex = base_color_prex;
			this.twisted = twisted;
			lock (next) {next = next_arg;}
		}

		/**
		 * Go to the next node. Since another thread may be changing this node,
		 * access must be mutexed.
		 */
		public Node* go_next () {
			lock (next) {
				return next;
			}
		}
		
		/**
		 * Claim a node. Return false if the node is already claimed by another thread.
		 */
		public bool claim () {
			lock (claimed) {
				if (!claimed) return claimed = true;
				else return false;
			}
		}

		/**
		 * To avoid waiting, a baton holder must give up its baton within a claimed node
		 * that is still being expanded by another thread
		 */
		bool drop_baton (ref bool baton) {
			lock (claimed) {
				if (claimed && !baton_passing) {
					baton = false;
					baton_passing = true;
					return true;
				}
				claimed = true;
				return false;
			}
		}

		/**
		 * If the baton holder has dropped the baton, pick it.
		 */
		bool pick_baton (ref bool baton) {
			lock (claimed) {
				if (baton_passing) {
					baton = true;
					return true;
				}
				else {
					if (!leaf_computed) {
						// This node is reused for its own left-most child.
						// Unclaim it to make it available to some other thread.
						claimed = false;
					}
					else {
						// Done with this node. Leave the finishing touch to the
						// baton holder.
						baton_passing = true;
					}
					return false;
				}
			}
		}

		/**
		 * Expand the node.
		 * A node is considered expanded if it is a leaf,
		 * or, for a non-leaf, its childs are generated.
		 */
		public void expand (ref bool baton) {
			if (baton && drop_baton (ref baton)) return;

			// The loop is for baton-holder.
			// Non-baton-holder perform this only once.
			do {
				if (!leaf_computed) compute_leaf (baton);
				if (!baton) {
					if (!leaf) {
						generate_childs ();
					}
					if (!pick_baton (ref baton)) return;
					// Now holding the baton
					if (!leaf_computed) continue;
					// else proceed below
				}
						
				// baton holder proceed here
				if (leaf) {
					Cairo.Context cr = fractalizer.cr;
					cr.set_line_width (width);
					cr.set_source_rgba (color.r, color.g, color.b, color.a);
					dc.render (cr);
				}
				else if (!expanded) {
					generate_childs ();
				}
			} while (!expanded);
		}

		/**
		 * Is this node a leaf?
		 * If so, then mark it as expanded.
		 * A non-baton holder should also create an image surface (by calling DeferredCairo.pixelize)
		 * so that the baton holder has less work to do (presumably copying an image surface
		 * is less work than rendering a multi-line shape).
		 */
		void compute_leaf (bool baton) {
			// Don't change the order of x||y below because we need the side effect to 'width'
			leaf = (seed.too_thin (this, fractalizer.probe, fractalizer.resolution, twisted, ref width)
				       || level_of_detail == 0);
			leaf_computed = true;

			if (leaf) {
				expanded = true;
				width = (width * 0.5).clamp (0.4, 1.0);
				// Mix base color
				color = RGBA (base_mix_cmp * color.r + base_color_prex.r,
					          base_mix_cmp * color.g + base_color_prex.g,
					          base_mix_cmp * color.b + base_color_prex.b,
					          base_mix_cmp * color.a + base_color_prex.a);
				seed.draw (dc, this, fractalizer.probe, fractalizer.delta);
				if (!baton) {
					dc.pixelize (width, color);
				}
			}
		}

		/**
		 * Generate child nodes.
		 * The left-most child might be generated "in place" using this node.
		 * Other side effect: expanded, leaf_computed, last_child
		 */
		void generate_childs () {
			Node* parent_link;
			var off_list_parent = depth > 2;
			if (!off_list_parent) {
				parent_link = this;
			}
			else {
				// Copy this node to a new off-list node. This node will then be reused
				// for its left-most child. The off-list node is used for the Projection part.
				// But, last_child and parent link is also copied for the purpose of garbage disposal.
				parent_link = Node.create ();
				parent_link->copy_projection (this);
				parent_link->parent = parent;
				parent_link->last_child = last_child;
			}
			// Pre-multiply base color
			if (seed.role == Role.BASE) {
				base_mix_cmp = 1 - seed.color_mix;
				base_color_prex = RGBA (seed.color_mix * seed.color.r,
				                        seed.color_mix * seed.color.g,
				                        seed.color_mix * seed.color.b,
				                        seed.color_mix * seed.color.a);
			}
			// Generate from the right-most to the left-most
			Node* link = next;   // No atomic access required here
			var count_child = 0;
			var next_depth = depth + 1;
			for (int n = fractalizer.pod.size - 1; n >= 0; n--) {
				var sub_seed = fractalizer.pod[n];
				if (depth == 0 && sub_seed.role == Role.MOTIF || 
					sub_seed.role == Role.BASE && depth >= 1) {  // Put depth>=1 last as it is less discriminatory
					continue;
				}
				// Next level of detail
				var next_level_of_detail = int.min (level_of_detail - 1, sub_seed.level_of_detail);
				// Mix current color with next seed color
				var u = (depth == 0 || seed.role == Role.BASE) ?
			              1.0 : Math.pow (sub_seed.color_mix, next_depth);
				var v = 1 - u;
				var color2 = RGBA (v * color.r + u * sub_seed.color.r,
					               v * color.g + u * sub_seed.color.g,
					               v * color.b + u * sub_seed.color.b,
					               v * color.a + u * sub_seed.color.a);
				for (int i=seed.num_segment - 1; i >= 0;  i--) {
					Node* node;
					if (i == 0 && n == fractalizer.first_non_base && off_list_parent) {
						// Reuse this node as its own left-most child
						node = this;
						// Don't count this node towards the tree_size
						last_child = (count_child == 0);
						// Treat this as a new node
						leaf_computed = false;
					}
					else {
						node = Node.create ();
						if (count_child++ == 0) node->last_child = true;
					}
					node->init (fractalizer, next_depth, seed.projection[i], parent_link, sub_seed,
						next_level_of_detail,
						color2, base_mix_cmp, base_color_prex,
						twisted || seed.check_twisted (i),
						link);
					link = node;
				}//for i
			}//for n
			if (!off_list_parent) {
				lock (next) {next = link;}
				expanded = true;
			}
			GLib.AtomicInt.add (ref fractalizer.tree_size, count_child);
		}
	}//class Node


	/**
	 * A pod defines a fractal using coordinates normalized to 0..1
	 * These coordinates will be projected to actual coordinates.
	 */
	weak Pod pod;

 	/**
	 * Index to the first non-BASE seed in the pod.
	 * This is useful because the left-most child of a Node is typically
	 * the first segment in the first non-BASE seed.
	 * Used by Node.generate_childs
	 */
	int first_non_base;

	/**
	 * Set by fractalize and fractalize_line.
	 * Used by Node
	 */
	weak Cairo.Context cr;

	/**
	 * The seed to be fractalized should have real device coordinates.
	 */
	weak Seed seed;
	bool quality_rendering;

	/**
	 * The fractalization process can be cancelled.
	 */
	weak Probe? probe;

	/**
	 * A widget can be periodically updated as a feedback to the user indicating the progress.
	 */
	weak IRender? widget;

	/**
	 * Rendering controls :
	 * resolution : passed to Seed.too_thin to check if more detailed rendering is required.
	 * delta : passed to Seed.draw
	 */
	double resolution;
	double delta;

	public static void draw (Pod pod, Cairo.Context cr, Seed seed, Probe? probe = null,
	                           IRender? widget = null, bool quality_rendering = true) {
		seed.start_render ();
		foreach (var s in pod) s.start_render ();
		
		var f = new Fractalizer ();
		f.resolution = (seed.width0 / 100).clamp (0.9, 4.0);
		f.delta = quality_rendering ? 0.5 * f.resolution : 5;
		f.pod = pod;
		f.seed = seed;
		f.probe = probe;
		f.widget = widget;
		f.quality_rendering = quality_rendering;

		// Find the first non-BASE seed.
		for (f.first_non_base = 0; f.first_non_base < f.pod.size; f.first_non_base++) {
			if (f.pod[f.first_non_base].role != Role.BASE) break;
		}

		cr.save ();
		f.fractalize (cr);
		cr.restore ();

		seed.stop_render ();
		foreach (var s in pod) s.stop_render ();
	}

	/**
	 * Thread pool.
	 */
	Thread<void*>[] threads = null;
	bool finished;

	/**
	 * True if we are rendering a line of constant width which
	 * can be optimized by rendering only one segment and replicating
	 * this segment.
	 *
	 * Note: we cannot simply replicate this segment in one go.
	 * Replication should be done bit by bit to ensure correct layering
	 * of overlapping patterns. We do this by triggering the replication
	 * when node.depth == 2.
	 */
	bool replicating;

	/**
	 * tree_size is used to prevent the tree from growing too large.
	 */
	int tree_size = 0;

	/**
	 * Start threads to grow a tree starting from the given root.
	 */
     [CCode (cname = "g_get_num_processors")]
     private extern static uint get_num_processors ();
	void start_threads (Node* root) {
		finished = false;
		// Don't multi-thread if there is only 1 processor
		var num_threads = uint.min (8, get_num_processors () - 1);   // Leave 1 for the main thread
		// Check environment for testing on single thread
		if (Environment.get_variable ("SINGLE_THREAD") != null) num_threads = 0;
		if (!quality_rendering) num_threads = 0;
		root->count_left_link = (int) num_threads + 1;
		if (num_threads > 0) {
			threads = new Thread<void*>[num_threads];
			for (int i = 0; i < num_threads; i++) {
				threads[i] = new Thread<void*> ("MutiaraThread", () => {run_thread (root, false); return null;});
			}
		}
		// The main thread also runs the same routine
		run_thread (root, true);
		close_threads ();
	}

	void close_threads () {
		if (threads == null) return;
		foreach (var thread in threads) {
			thread.join ();
		}
		threads = null;
	}

	/**
	 * Entry point for thread
	 */
	void run_thread (Node* root, bool baton) {
		Node* left_most = root;
		while (!finished) {
			var node = left_most;  // The left-most unexpanded node
			if (baton) flush (node);
			if (node == null) break;
			var all_expanded = true;    // false after encountering a node that has not been expanded
			if (baton) {
				// A baton holder must visit the node even if it is already claimed
				node->expand (ref baton);
			}
			else {
				// Release expanded left nodes before checking for access. Otherwise,
				// proper garbage disposal will be prevented by threads that sleep and loop
				// waiting for tree_size to become small.
				var next_left_most = left_most;
				while (next_left_most != null && next_left_most->expanded) {
					next_left_most = next_left_most->go_next ();
				}
				if (left_most != next_left_most) advance_pointer (&left_most, next_left_most);
				node = left_most;
				// To prevent excessive memory usage, a non-baton holder cannot increase
				// the tree size beyond a limit.
				int excess;
				if ((excess = AtomicInt.@get (ref tree_size) - 5000) > 0) {
					Thread.usleep (excess);
					continue;
				}
				// A non-baton holder must search for an unclaimed node
				while (node != null && !node->claim ()) {
					if (all_expanded) {
						if (!node->expanded) {
							all_expanded = false;
							next_left_most = node;
						}
					}
					node = node->go_next ();
				}
				if (left_most != next_left_most) advance_pointer (&left_most, next_left_most);
				if (node != null) node->expand (ref baton);
				else if (all_expanded) break;     // No more nodes to be expanded
			}
			if (baton) {
				if (probe != null && probe.cancelled) break;
				advance_pointer (&left_most, node->go_next ());
			}
		}
		if (baton) {
			finished = true;
		}
		// Clean up.
		advance_pointer (&left_most, null);
	}

	/**
	 * Advanced left_most pointer, reclaming any garbage (and their parents)
	 * between the previous left_most and the new left_most.
	 *
	 * This scheme has been verified to be leak-free.
	 */
	void advance_pointer (Node** left_most, Node* new_left_most) {
		if (new_left_most != null) AtomicInt.inc (ref new_left_most->count_left_link);
		Node* garbage = *left_most;
		*left_most = new_left_most;
		while (garbage != null && AtomicInt.dec_and_test (ref garbage->count_left_link)) {
			// No other threads can enter this. Hence, there is no need
			// for atomic access to anything.
			// [Don't delete this comment lest I forget and do stupid things]
			// Can dispose be optimized given that only one thread can do that
			// at any time? Answer: No. There might be another thread from another
			// call to fractalizer.
			Node *next = garbage->go_next ();
			// The garbage can be deleted if it is not a parent of any projection.
			// i.e., if it is a leaf node, or not expanded (this later case happens
			// when the rendition is cancelled).
			if (garbage->leaf || !garbage->expanded) {
				do {
					var not_the_last_child = !garbage->last_child;
					Node *up = garbage->parent;
					Node.dispose (garbage);
					if (not_the_last_child) break;
					// if garbage is last child, dispose its parent as well.
					garbage = up;
				} while (garbage != null);
			}
			GLib.AtomicInt.add (ref tree_size, -1);
			garbage = next;
		}
	}
	
	/**
	 * "Flush" the current rendition
	 */
	void flush (Node? node) {
		if (replicating && (node == null || node.depth == 2)) {
			replicate_short_line ();
		}
		// Update visually on the screen
		if (widget != null && (node == null || probe != null && probe.time_to_peek ())) {
			widget.render (node == null ? 1.0 : 0.5);
		}
	}

	/**
	 * Fractalize the given seed.
	 */
	void fractalize (Cairo.Context cr) {
		if (seed is Line && Math.fabs(seed.width0 - seed.width1) < 1.0) {
			fractalize_line (cr);
			return;
		}

		this.cr = cr;

		Node* root = Node.create_root (this, seed);
		start_threads (root);
	}

	/**
	 * The following are used for optimized rendition of a line of constant width.
	 * See the comment on "replicating" variable.
	 */
	Seed short_line;
	double short_line_length;
	Cairo.ImageSurface short_line_image;
	Cairo.Context short_line_cr;
	Cairo.Context long_line_cr;

	/**
	 * Optimized rendering of line of constant width
	 */
	void fractalize_line (Cairo.Context long_line_cr) {
		replicating = true;
		this.long_line_cr = long_line_cr;
		// Create a short segment that has all the characteristics of the full line
		// except for its length, orientation (horizontal) and base offset (0)
		var width = seed.width0;
		var p0 = seed.point (0);
		var p1 = seed.point (1);
		var length = Point.hypot (p0, p1);
		short_line_length = length / seed.num_segment;
		short_line = new Line ();
		short_line.copy_parameters (seed);
		short_line.init_full ({short_line_length, width, short_line_length * 2, width});
		short_line.num_segment = 1;
		short_line.base_offset = 0;
		short_line.start_render ();
		short_line_image = new Cairo.ImageSurface (Cairo.Format.ARGB32, (int)(short_line_length * 3), (int)(width * 3));
		short_line_cr = new Cairo.Context (short_line_image);
		cr = short_line_cr;
		
		// Rotate to make the main line horizontal. This makes copying easier (really??)
		long_line_cr.translate (p0.x, p0.y);
		long_line_cr.rotate (Math.atan2 (p1.y - p0.y, p1.x - p0.x));
		long_line_cr.translate (- short_line_length, (-1 - seed.base_offset) * width);

		Node *root = Node.create_root (this, short_line);
		start_threads (root);
	}

	/**
	 * Replicate short_line_image onto the full line
	 */
	void replicate_short_line () {
		var width = seed.width0;
		for (int i=0; i < seed.num_segment; i++) {
			long_line_cr.save ();
			long_line_cr.translate (i * short_line_length, 0);
			if (i % 2 == 1) {
				if (seed.flip_even.y () != seed.flip_odd.y ()) {
					long_line_cr.translate (0, 1.5 * width);
					long_line_cr.scale (1, -1);
					long_line_cr.translate (0, -1.5 * width);
				}
				if (seed.flip_even.x () != seed.flip_odd.x ()) {
					long_line_cr.translate (1.5 * short_line_length, 0);
					long_line_cr.scale (-1, 1);
					long_line_cr.translate (-1.5 * short_line_length, 0);
				}
			}
			long_line_cr.set_source_surface (short_line_image, 0, 0);
			long_line_cr.paint ();
			long_line_cr.restore ();
		}
		// Clear short_line_image
		cr.set_operator (Cairo.Operator.CLEAR);
		cr.paint ();
		cr.set_operator (Cairo.Operator.OVER);
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
