/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
public class Line : Seed {

    /**
     * The following variables are made public for the benefit of
     * Projection class, which may merge two or more affine transformations.
     */
	public Point p0;
	public Point p1;
	public double sin_perpendicular_angle;
	public double cos_perpendicular_angle;

	protected override void init_default_sub (Point p) {
		p0 = p;
		p1 = p;
		memo_perpendicular_angle ();
	}

	protected override void init_full_sub (double[] args) {
		assert (args.length >= 4);
		p0 = Point () {x = args[0], y = args[1]};
		p1 = Point () {x = args[2], y = args[3]};
		memo_perpendicular_angle ();
	}

	protected override string serialize_sub () {
		return "Line (%s, %s) (%s, %s)".printf (
			Util.format (p0.x, 4), Util.format (p0.y, 4),
			Util.format (p1.x, 4), Util.format (p1.y, 4));
	}

	protected override Point point_raw (double t) {
		// (1-t) p0 + t p1
		var s = 1.0 - t;
		return Point () { x = s * p0.x + t * p1.x, y = s * p0.y + t * p1.y};
	}

	protected override void init_distance_to_parametric () {
		memo_perpendicular_angle ();
		is_affine = Math.fabs (width0 - width1) < 0.0000001;
	}

	public override Point distant_point (double x_distance, double y_distance) {
		// scale y_distance by the width at x_distance
		y_distance *= ((1 - x_distance) * width0  +  x_distance * width1);

		// point at x_distance
		//    .move by y_distance along the perpendicular
		var t = x_distance;   // For Line, parameter (x_distance) equals x_distance
		var s = 1.0 - t;
		return Point () {
			x = (s * p0.x + t * p1.x)  +  y_distance * cos_perpendicular_angle,
			y = (s * p0.y + t * p1.y)  +  y_distance * sin_perpendicular_angle};
	}

	void memo_perpendicular_angle () {
		Math.sincos (Math.atan2 (p0.x - p1.x, p1.y - p0.y),
			out sin_perpendicular_angle,
			out cos_perpendicular_angle);		
	}

	public override void draw_direct (Cairo.Context cr, double scale) {
		Point c0 = p0 .scale (scale, scale);
		Point c1 = p1 .scale (scale, scale);
		cr.move_to (c0.x, c0.y);
		cr.line_to (c1.x, c1.y);
		cr.stroke ();
	}

	public override void draw_controls (Cairo.Context cr, double scale) {
		draw_square_control (cr, p0 .scale (scale, scale));
		draw_square_control (cr, p1 .scale (scale, scale));
	}

	/**
	 * Control:
	 *  0, 1 : end point
	 *  100  : the line
	 */
	Point p0_save;
	Point p1_save;

	protected override int pressed_control (Point p, double scale) {
		return within_touch_range (p, p1, scale) ? 1 :
			   within_touch_range (p, p0, scale) ? 0 : -1;
	}

	protected override void save_drag_context () {
		p0_save = p0;
		p1_save = p1;
	}

	protected override void drag_control (Point p, double scale, bool snap_on, double grid_size, double snap_angle) {
		if (snap_on) snap (ref p, scale, grid_size);
		if (selected_control == 0) p0 = p;
		else if (selected_control == 1) p1 = p;
		else assert_not_reached ();
	}

	protected override void move_shape (double delta_x, double delta_y) {
		p0 = p0_save.move_xy (delta_x, delta_y);
		p1 = p1_save.move_xy (delta_x, delta_y);
	}

	protected override Point[] points_to_snap () {
		return {p0, p1};
	}

	public override void draw_icon (Cairo.Context cr, double width, double height) {
		double[] args = {0, 0, width, height};
		init_full (args);
		draw_direct (cr, 1);
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
