/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
using Gee;


public class MotifChooser : Gtk.DrawingArea, IRender {
	public signal void changed (Pod motif);
	public bool dark_background {get; set;}

	ArrayList<Pod> motifs = new ArrayList<Pod> ();
	int selected = -1;

	/**
	 * Screen dimensions for the selection images
	 */
	static int size = 30;
	static int gap = 5;
	static int preview_length = 4;
	static int preview_width = 120;  // preview_length * size
	static int preview_height = 40;  // gap + size + gap;
	static int motifs_per_row;

	ArrayList<RenderJob> render_jobs = new ArrayList<RenderJob> ();

	public MotifChooser () {
		can_focus = true;
		tooltip_text = _("Double-click to select");
		add_events (Gdk.EventMask.BUTTON_PRESS_MASK);
		draw.connect (draw_motifs);
		button_press_event.connect ((s, e) => {
			if (e.button == 1) {
				if (e.type == Gdk.EventType.@2BUTTON_PRESS) double_click (e.x, e.y);
				else click (e.x, e.y); 
			}
			return true; });
		key_press_event.connect ((s, e) => {
			// TODO
			return false;
		});
		realize.connect ((s) => {
			add_inline_motifs.begin ();
			parent.size_allocate.connect (() => {queue_resize ();});
		});
		notify["dark-background"].connect ((s, p) => {
			queue_draw ();
		});
	}

	/**
	 * Height for width geometry
	 */
	public override Gtk.SizeRequestMode get_request_mode () {
		return Gtk.SizeRequestMode.HEIGHT_FOR_WIDTH;
	}

	public override void get_preferred_width (out int min_width, out int ideal_width) {
		var width = preview_width + gap + gap;
		min_width = width;
		ideal_width = width;
	}

	public override void get_preferred_height_for_width (int width, out int min_height, out int ideal_height) {
		motifs_per_row = (int) (width / (preview_width + gap + gap));
		if (motifs_per_row == 0) motifs_per_row = 1;
		var num_rows = motifs.size / motifs_per_row + 1;
		var height = preview_height * num_rows;
		min_height = height;
		ideal_height = height;
	}

	/**
	 * Add a motif to the selection. Create a private copy of the motif.
	 */
	public new void add (Pod pod) {
		// Check for duplicate
		foreach (var motif in motifs) {
			if (motif.same_as (pod)) return;
		}
		var motif = new Pod ();
		motif.copy (pod);
		motifs.add (motif);
		render_jobs.add (create_render_job ());
		queue_resize ();
		assert (render_jobs.size == motifs.size);
	}

	bool draw_motifs (Cairo.Context cr) {
		var color = (dark_background ? 0.0 : 1.0);
		cr.set_source_rgb (color, color, color);
		cr.paint ();
		// Get clipping region to minimize work
		double clip_x1;
		double clip_y1;
		double clip_x2;
		double clip_y2;
		cr.clip_extents (out clip_x1, out clip_y1, out clip_x2, out clip_y2);
		clip_x1 -= preview_width;    // Partially visible motifs must also be drawn
		clip_y1 -= preview_height;
		// Highlight selected motif
		if (selected != -1) {
			var row = selected / motifs_per_row;
			var col = selected % motifs_per_row;
			cr.rectangle (col * (preview_width + gap + gap) + gap, row * preview_height, preview_width, preview_height);
			cr.set_source_rgba (0.5, 0.5, 0.5, 0.4);
			cr.fill ();
		}
		for (var n=0; n < motifs.size; n++) {
			cr.save ();
			var row = n / motifs_per_row;
			var col = n % motifs_per_row;
			var x = col * (preview_width + gap + gap) + gap;
			var y = row * preview_height;
			if (x >= clip_x1 && y >= clip_y1 && x <= clip_x2 && y <= clip_y2) {
				render_jobs[n].paint_image_surface (cr, x, y);
			}
			cr.restore ();
		}
		return true;
	}

	public void render (double fraction) {
		queue_draw ();
	}

	/**
	 * Create a render job for the most recently added motif
	 */
	RenderJob create_render_job () {
		var motif = motifs.last ();
		var line = new Line ();
		line.init_full ({-size, gap, preview_width + size, gap});
		line.base_offset = 0;
		line.width0 = line.width1 = size;
		line.num_segment = preview_length + 2;   // Add 2 to fill in the left and right most
		line.flip_even = motif.flip_even;
		line.flip_odd = motif.flip_odd;
		line.tweak = motif.tweak;
		line.level_of_detail = 100;
		var pod = new Border (motif);
		pod.add (line);
		return RenderJob.create_named ("gallery %d".printf (render_jobs.size), pod, 0, 0, preview_width, preview_height, this);
	}

	void click (double x, double y) {
		grab_focus ();
	}

	void double_click (double x, double y) {
		var row = (int) (y / preview_height);
		var col = (int) (x / (preview_width + gap + gap));
		var n = row * motifs_per_row + col;
		if (n >= 0 && n < motifs.size) {
			selected = n;
		}
		else {
			selected = -1;
		}
		if (selected != -1) {
			changed (motifs[selected]);
			queue_draw ();
		}
	}

	async void add_inline_motifs () {
		var ss = new Gallery ().motifs.split ("\n-----\n");
		foreach (var s in ss) {
			var motif = new Pod ();
			motif.compile (s);
			if (motif.test && !Util.debug_on ()) continue;
			add (motif);
			Idle.add (add_inline_motifs.callback);
			yield;
		}
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
