/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
using Gee;

public class Pod : Object {
	public signal void signal_changed ();

	/**
	 * True if the pattern is designed to test Mutiara
	 * and not intended for the user.
	 */
	public bool test = false;

	/**
	 * Symmetry type.
	 */
	public Flip flip_even;
	public Flip flip_odd;
	public Tweak tweak;
	ArrayList<Seed> seeds = new ArrayList<Seed> ();

	/**
	 * The level of fractalization. 0 means non-fractal.
	 */
	public int level_of_detail;

	public Pod () {
		set_defaults ();
	}

	/**
	 * Don't change these default. It is assumed in serialized version.
	 */
	void set_defaults () {
		flip_even = Flip.NONE;
		flip_odd = Flip.Y;
		tweak = Tweak ();
		level_of_detail = 100;
	}

	/**
	 * Convert to text that can be compiled ().
	 */
	public string serialize () {
		string? locale = Intl.setlocale (LocaleCategory.NUMERIC, null);
		Intl.setlocale (LocaleCategory.NUMERIC, "C");
		string text = serialize_aux ();
		Intl.setlocale (LocaleCategory.NUMERIC, locale);
		return text;
	}

	protected virtual string serialize_aux () {
		var s = "";
		if (level_of_detail != 100) s += "level-of-detail %d\n".printf (level_of_detail);
		if (flip_even != Flip.NONE || flip_odd != Flip.Y) s += "flip %d %d\n".printf (flip_even, flip_odd);
		if (!tweak.equals (Tweak ())) s += "tweakwhole %s\n".printf (tweak.serialize());
		foreach (var seed in seeds) s += seed.serialize ();
		return s;
	}

	/**
	 * Quick and dirty way of copying
	 */
	public void copy (Pod pod) {
		compile (pod.serialize ());
	}

	/**
	 * Check if this pod is approximately equal to another
	 */
	public bool same_as (Pod pod) {
		return serialize () == pod.serialize ();
	}

	public static int search (ArrayList<Pod> pods, Pod pod) {
		for (int i=0; i < pods.size; i++) {
			if (pods[i].same_as (pod)) return i;
		}
		return -1;
	}

	/**
	 * A very forgiving compiler
	 */
	public virtual void compile (string s) {
		set_defaults ();
		seeds.clear ();
		// The initial seed is just a dummy to stop any null pointer error
		Seed null_seed = new Line ();
		Seed seed = null_seed;
		foreach (var line in s.split ("\n")) {
			// line --> command arguments
			var cmdarg = line.strip ().split (" ", 2);
			if (cmdarg.length == 0) continue;
			double[] arg = {};
			if (cmdarg.length == 2) {
				foreach (var token in Regex.split_simple ("[\\(|\\)|\\,|\\s]++", cmdarg[1])) {
					if (token.length == 0) continue;
					arg += double.parse (token);
				}
				// Other arguments default to 0.0
				for (int i=0; i < 10; i++) arg += 0.0;
			}
			switch (cmdarg[0]) {
				case "Line" :
				case "Curve" :
				case "Arc" :
					seed = (Seed) Object.new (Type.from_name (cmdarg[0]));
					seed.init_full (arg);
					seeds.add (seed);
					break;
				case "test" :
					test = true;
					break;
				case "segmentation" :
					seed.base_offset = arg[0];
					seed.num_segment = (int) arg[1];
					break;
				case "width" :
					seed.width0 = arg[0];
					seed.width1 = arg[1];
					break;
				case "replication" :
					seed.role = (Role) arg[0];
					seed.level_of_detail = (int) arg[1];
					seed.flip_even = (Flip) arg[2];
					seed.flip_odd = (Flip) arg[3];
					break;
				case "level-of-detail" :
					level_of_detail = (int) arg[0];
					break;
				case "flip" :
					flip_even = (Flip) arg[0];
					flip_odd = (Flip) arg[1];
					break;
				case "tweak" :
					seed.tweak.scale_x = arg[0];
					seed.tweak.scale_y = arg[1];
					seed.tweak.rotate = arg[2];
					seed.tweak.move_x = arg[3];
					seed.tweak.move_y = arg[4];
					break;
				case "tweakwhole" :
					tweak.scale_x = arg[0];
					tweak.scale_y = arg[1];
					tweak.rotate = arg[2];
					tweak.move_x = arg[3];
					tweak.move_y = arg[4];
					break;
				case "color" :
					seed.color = RGBA (arg[0], arg[1], arg[2], arg[3]);
					seed.color_mix = arg[4];
					break;
				case "style" :
					seed.render_style = (RenderStyle) arg[0];
					break;
				case "debug" :
					seed.debug = true;
					break;
				case "#" :
					break;
				default:
					GLib.warning ("Unknown option: %s", cmdarg[0]);
					break;
			}
		}
		signal_changed ();
	}

	public virtual Seed? add_seed_at_point (Type type, Point p) {
		Seed? seed = Object.new (type) as Seed;
		assert (seed != null);
		seed.init_default (p);
		seeds.add (seed);
		return seed;
	}

	public Iterator<Seed> iterator () {
		return seeds.iterator ();
	}

	/**
	 * Access to seeds
	 */
	public int size {
		get {return seeds.size;}
	}

	public new Seed get (int n) {
		return seeds[n];
	}

	public int index_of (Seed? seed) {
		return seeds.index_of (seed);
	}

	public virtual void remove (Seed seed) {
		seeds.remove (seed);
	}

	public virtual void add (Seed seed) {
		seeds.add (seed);
	}

	public virtual void clear () {
		seeds.clear ();
	}

	/**
	 * Swap seed positions
	 */
	public virtual void swap (int i, int j) {
		var t = seeds[i];
		seeds[i] = seeds[j];
		seeds[j] = t;
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
