/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
using Gee;

public class PodDrawing : Gtk.Notebook, IRender {
	/**
	 * The pod to be drawn.
	 */
	public Border base_pod;
	/**
	 * The pod to be edited
	 */
	public Pod edited_pod;

	/**
	 * Is the panel hidden?
	 * Don't render if true.
	 */
	public bool hidden {get; set;}

	/**
	 * True if edited_pod is changed while the panel is hidden.
	 */
	bool stale;

	/**
	 * Drawing area
	 */
	public Gtk.DrawingArea drawing_area;

	/**
	 * Zoom
	 */
	Zoom zoom_control = new Zoom ();
	double zoom;

	/**
	 * Grid and snap.
	 */
	public double grid_size {get; set;}
	public double snap_angle;  // In degree rather than radian

	/**
	 * All properties of the form *-visible causes drawing_area to
	 * be redrawn upon change notification.
	 */

	/**
	 * false : only selected visible
	 * true : all visible
	 */
	public bool all_seeds_visible {get; set;}

	public bool grid_visible {get; set;}
	public bool segmentation_visible {get; set;}
	public bool rendering_visible {get; set;}
	public bool dark_background {get; set;}

	/**
	 * Either 1 for white or 0 for black
	 */
	double background;

	/**
	 * total_width and total_height includes borders.
	 */
	int origin_x;
	int origin_y;
	int total_width = 0;
	int total_height;
	int gap = 100;

	/**
	 * Editing area.
	 * (pod_x, pod_y) is relative to (origin_x, origin_y)
	 * pod_scale is 1 for border editing, and the view size for motif editing.
	 */
	int pod_x;
	int pod_y;
	public int pod_width {get; private set;}
	public int pod_height {get; private set;}
	double pod_scale;

	/**
	 * Default values for newly created seed
	 */
	public Seed default_seed;

	/**
	 * The state of the interaction.
	 */
	public Seed selected_prototype = null;
	public Seed selected_seed {get; set;}  // public, so that its changes can be notified
	bool dragging = false;

	/**
	 * The centre of the visible area
	 */
	double centre_x;
	double centre_y;
	double visible_width;
	double visible_height;

	/**
	 * Support for drag scrolling
	 */
	bool scrolling = false;
	int pressed_x;
	int pressed_y;
	double save_centre_x;
	double save_centre_y;

	/**
	 * Scroll bars
	 */
	Gtk.Adjustment h_adjustment;
	Gtk.Adjustment v_adjustment;

	/**
	 * Don't invoke queue_draw when draw_requested is already true
	 */
	bool draw_requested;

	/**
	 * Area for textual editing
	 */
	Gtk.TextView text_area;

	/**
	 * Async rendering
	 */
	RenderJob render_job;
	bool use_shared_render_job;

	/**
	 * Widget to contain progress bars.
	 * Set by PodEditor
	 */
	public Gtk.Grid status_box;

	public PodDrawing.border (Border border) {
		resize (640, 480);
		pod_scale = 1.0;
		grid_size = 100.0;
		// Default values for newly created seed
		var seed = new Line ();
		seed.level_of_detail = 100;
		seed.width0 = 100;
		seed.width1 = 100;
		use_shared_render_job = true;
		common_init (border, border, seed);
	}

	public PodDrawing.motif (Pod motif) {
		var motif_scale = 400;
		pod_x = motif_scale;
		pod_y = 0;
		pod_width = motif_scale;
		pod_height = motif_scale;
		origin_x = gap;
		origin_y = gap;
		total_width = gap + pod_width * 3 + gap;
		total_height = gap + pod_height + gap;
		pod_scale = motif_scale;
		grid_size = 0.1;
		// Create a pod to preview the motif
		var preview_pod = new Border (motif);
		var line = new Line ();
		var y = 0.5 * motif_scale;
		var x = 0;
		line.init_full ({x, y, x + 3 * motif_scale, y});
		line.width0 = line.width1 = motif_scale;
		line.flip_even = preview_pod.flip_odd;
		line.flip_odd = preview_pod.flip_even;
		line.tweak = preview_pod.tweak;
		line.level_of_detail = 100;
		line.num_segment = 3;
		preview_pod.add (line);
		// Default values for newly created seed
		var seed = new Line ();
		seed.width0 = 0.2;
		seed.width1 = 0.2;
		common_init (preview_pod, motif, seed);
	}

	void common_init (Border base_pod, Pod edited_pod, Seed default_seed) {
		this.base_pod = base_pod;
		this.edited_pod = edited_pod;
		this.default_seed = default_seed;
		hidden = true;
		all_seeds_visible = true;
		segmentation_visible = false;
		rendering_visible = true;
		grid_visible = true;
		centre_x = 0.5 * total_width;
		centre_y = 0.5 * total_height;
		snap_angle = 15.0;

		// Layout
		this.show_tabs = false;
		this.append_page (create_drawing_area (), new Gtk.Label ("Drawing"));
		this.append_page (create_text_area (), new Gtk.Label ("Text"));

		// Signals
		edited_pod.signal_changed.connect (() => {
			stale = true;
			create_image_surface_lazily ();
			// This may occur when the pod is completely replaced
			if (edited_pod.index_of (selected_seed) == -1) selected_seed = null;
		});
		this.notify.connect ((s, p) => {
			var name = p.name;
			if (name == "selected-seed" || name == "grid-size" ||
			    name == "dark-background" || name.has_suffix ("-visible")) {
				background = (dark_background ? 0.0 : 1.0);
				redraw ();
			}
			else if (name == "hidden") {
				create_image_surface_lazily ();
			}
		});
		zoom = 1.0;
		zoom_control.signal_changed.connect ((z) => {
			zoom = z;
			change_adjustments ();
			redraw ();
		});
		drawing_area.can_focus = true;
		drawing_area.key_press_event.connect ((s, e) => {return handle_key_press (e.keyval);});
		selected_seed = null;
	}

	/**
	 * Interpret key press
	 * TODO: keys to nudge seeds
	 */
	bool handle_key_press (uint keyval) {
//		if (keyval == Gdk.keyval_from_name ("Delete")) {  // Delete is already hot-keyed
//			delete_selection ();
//		}
//		else {
//			return false;
//		}
		return false;
	}

	/**
	 * Make the flipness of the preview pod consistent with the default setting
	 */
	public void flip_preview_pod () {
		foreach (var s in base_pod) {
			s.flip_even = edited_pod.flip_odd;
			s.flip_odd = edited_pod.flip_even;
			s.tweak = edited_pod.tweak;
		}
		edited_pod.signal_changed ();   // Force the preview to be redrawn
	}

	/**
	 * This is applicable to PodDrawing.border only
	 */
	public void resize (int width, int height) {
		pod_x = 0;
		pod_y = 0;
		pod_width = width;
		pod_height = height;
		origin_x = gap;
		origin_y = gap;
		total_width = gap + pod_width + gap;
		total_height = gap + pod_height + gap;
		if (drawing_area != null) {
			create_image_surface ();
		}
		change_adjustments ();
	}

	/**
	 * Adjust scrollbar settings
	 */
	void change_adjustments () {
		if (h_adjustment == null) return;
		h_adjustment.upper = total_width + visible_width / zoom;
		h_adjustment.page_size = visible_width / zoom;
		h_adjustment.page_increment = visible_width / zoom / 2;
		h_adjustment.value = centre_x;

		v_adjustment.upper = total_height + visible_height / zoom;
		v_adjustment.page_size = visible_height / zoom;
		v_adjustment.page_increment = visible_height / zoom / 2;
		v_adjustment.value = centre_y;
	}

	public void select_previous () {
		drawing_area.grab_focus ();
		var size = edited_pod.size;
		if (selected_seed != null) {
			var index = edited_pod.index_of (selected_seed);
			selected_seed = edited_pod[int.max (0, index - 1)];
		}
		else if (size > 0) {
			selected_seed = edited_pod[size - 1];
		}
	}

	public void select_next () {
		drawing_area.grab_focus ();
		var size = edited_pod.size;
		if (selected_seed != null) {
			var index = edited_pod.index_of (selected_seed);
			selected_seed = edited_pod[int.min (size - 1, index + 1)];
		}
		else if (size > 0) {
			selected_seed = edited_pod[0];
		}
	}

	/**
	 * move_selection_* is broken up to avoid multiple signals
	 */
	void move_selection_up_aux () {
		var index = edited_pod.index_of (selected_seed);
		if (index != -1 && index != edited_pod.size - 1) swap_selected_seed (index + 1);
	}

	void move_selection_down_aux () {
		var index = edited_pod.index_of (selected_seed);
		if (index > 0) swap_selected_seed (index - 1);
	}

	void end_move_selection () {
		edited_pod.signal_changed ();
	}

	public void move_selection_up () {
		move_selection_up_aux ();
		end_move_selection ();
	}

	public void move_selection_down () {
		move_selection_down_aux ();
		end_move_selection ();
	}

	public void move_selection_to_top () {
		for (var i = 0; i < edited_pod.size; i++)
			move_selection_up_aux ();
		end_move_selection ();
	}

	public void move_selection_to_bottom () {
		for (var i = 0; i < edited_pod.size; i++)
			move_selection_down_aux ();
		end_move_selection ();
	}

	/**
	 * Move the selected seed to position n within its pod
	 */
	void swap_selected_seed (int n) {
		var index = edited_pod.index_of (selected_seed);
		if (index == -1 || index == n) return;
		edited_pod.swap (index, n);
	}

	public void delete_selection () {
		if (selected_seed != null) {
			edited_pod.remove (selected_seed);
			selected_seed = null;
			edited_pod.signal_changed ();
		}
	}

	/**
	 * Create a scrollable drawing area and establish signal handlers for it.
	 */
	Gtk.Widget create_drawing_area () {
		h_adjustment = new Gtk.Adjustment (centre_x, 0, 0, 1, 1, 1);
		v_adjustment = new Gtk.Adjustment (centre_y, 0, 0, 1, 1, 1);
		var box = new Gtk.Grid ();
		box.attach (drawing_area = new Gtk.DrawingArea (), 0, 0, 1, 1);
		drawing_area.expand = true;
		box.attach (new Gtk.Scrollbar (Gtk.Orientation.HORIZONTAL, h_adjustment), 0, 1, 1, 1);
		box.attach (new Gtk.Scrollbar (Gtk.Orientation.VERTICAL, v_adjustment), 1, 0, 1, 1);
		drawing_area.add_events (Gdk.EventMask.POINTER_MOTION_MASK);
		drawing_area.add_events (Gdk.EventMask.BUTTON_MOTION_MASK);
		drawing_area.add_events (Gdk.EventMask.BUTTON_PRESS_MASK);
		drawing_area.add_events (Gdk.EventMask.BUTTON_RELEASE_MASK);
		drawing_area.draw.connect (draw_drawing_area);
		drawing_area.size_allocate.connect ((r) => {
			visible_width = r.width;
			visible_height = r.height;
			change_adjustments (); });
		h_adjustment.value_changed.connect (() => {
			centre_x = h_adjustment.value;
			redraw (); });
		v_adjustment.value_changed.connect (() => {
			centre_y = v_adjustment.value;
			redraw (); });
		drawing_area.button_press_event.connect ((s, e) => {
			if (e.button == 1) press (e.x, e.y, e.state);
			return true; });
		drawing_area.button_release_event.connect ((s, e) => {
			if (e.button == 1) release (e.x, e.y);
			return true; });
		drawing_area.motion_notify_event.connect ((s, e) => {
			move (e.x, e.y, e.state);
			return true; });
		return box;
	}

	/**
	 * Convenient interface to create shared rendering job
	 */
	public RenderJob create_shared_job (IRender? watch) {
		return RenderJob.create_shared (base_pod, origin_x, origin_y, total_width, total_height, watch);
	}

	void create_image_surface_lazily () {
		if (!hidden && stale) {
			create_image_surface ();
			stale = false;
		}
	}

	/**
	 * Create a cairo surface that is rendered asynchronously
	 */
	void create_image_surface () {
		// Cancel on-going async image generation
		if (use_shared_render_job && render_job != null) render_job.cancel (this);
		render_job = use_shared_render_job ?
			create_shared_job (this) :
			RenderJob.create_named ("editor", base_pod, origin_x, origin_y, total_width, total_height, this);
	}

	/**
	 * IRender implementation
	 */
	public void render (double fraction) {
		redraw ();
	}

	public async void save_image_surface (string filename) {
		var progress_box = new ProgressBox (status_box, _("Exporting"));
		var job = create_shared_job (progress_box);
		while (!job.image_surface_completed) {
			Timeout.add_seconds (2, save_image_surface.callback);
			yield;
		}
		// Copy surface from async job to surface with proper size
		var image_surface = new Cairo.ImageSurface (Cairo.Format.ARGB32, pod_width, pod_height);
		var cr = new Cairo.Context (image_surface);
		job.paint_image_surface (cr, -origin_x, -origin_y);
		image_surface.write_to_png (filename);
		job.cancel (progress_box);
		progress_box.remove ();
	}

	public async void print (Gtk.PrintOperation print_operation, Gtk.PrintContext ctx) {
		var progress_box = new ProgressBox (status_box, _("Printing"));
		var job = create_shared_job (progress_box);
		print_operation.set_defer_drawing ();
		while (!job.image_surface_completed) {
			Timeout.add_seconds (2, print.callback);
			yield;
		}
		// Copy surface from async job to printer
		var cr = ctx.get_cairo_context ();
		cr.scale (72.0 / 300, 72.0 / 300);
		job.paint_image_surface (cr, -origin_x, -origin_y);
		job.cancel (progress_box);
		progress_box.remove ();
		print_operation.draw_page_finish ();
	}

	void apply_zoom (Cairo.Context cr) {
		cr.scale (zoom, zoom);
		// Translate to int coordinates to ensure consistent "crispiness"
		cr.translate ((int) (visible_width / 2 / zoom - centre_x), (int) (visible_height / 2 / zoom - centre_y));
	}

	/**
	 * Translate to pod origin, but retain scaling at 1.0
	 * This is used to display instruments such as grids, segmentations, etc
	 * where the line thickness should not be affected by zooming.
	 */
	void translate_to_pod_origin (Cairo.Context cr) {
		apply_zoom (cr);
		cr.translate (origin_x + pod_x, origin_y + pod_y);
		cr.scale (1 / zoom, 1 / zoom);
	}

	/**
	 * Convert from screen coordinate to image coordinate relative to origin_x.
	 */
	double image_x (double x) {
		return centre_x + (x - visible_width / 2) / zoom - origin_x;
	}

	/**
	 * Convert from screen coordinate to image coordinate relative to origin_y.
	 */
	double image_y (double y) {
		return centre_y + (y - visible_height / 2) / zoom  - origin_y;
	}

	/**
	 * Redraw drawing_area
	 */
	void redraw () {
		if (!draw_requested) {
			draw_requested = true;
			drawing_area.queue_draw ();
		}
	}

	bool draw_drawing_area (Cairo.Context cr) {
		cr.set_source_rgb (background, background, background);
		cr.paint ();
		draw_frame (cr);
		draw_grid (cr);
		draw_rendering (cr);
		draw_segmentation (cr);
		draw_seeds (cr);
		zoom_control.draw (cr, background);
		draw_requested = false;
		return true;
	}

	/**
	 * Draw a frame around the editing area
	 */
	void draw_frame (Cairo.Context cr) {
		cr.save ();
		translate_to_pod_origin (cr);
		cr.set_line_width (1.0);
		cr.set_source_rgb (0.4, 0.4, 0.4);
		cr.rectangle (0, 0, pod_width * zoom, pod_height * zoom);
		cr.stroke ();
		cr.restore ();
	}

	void draw_grid (Cairo.Context cr) {
		if (!grid_visible) return;
		cr.save ();
		translate_to_pod_origin (cr);
		cr.set_line_width (0.5);
		cr.set_source_rgb (0.4, 0.4, 0.4);
		cr.set_dash ({1,2}, 0);
		for (var g = grid_size; g * pod_scale < pod_width; g += grid_size) {
			var x = g * zoom * pod_scale;
			cr.move_to (x, 0);
			cr.line_to (x, pod_height * zoom);
		}
		for (var g = grid_size; g * pod_scale < pod_height; g += grid_size) {
			var y = g * zoom * pod_scale;
			cr.move_to (0, y);
			cr.line_to (pod_width * zoom, y);
		}
		cr.stroke ();
		cr.restore ();
	}

	/**
	 * If high quality rendering is not finished yet, show a low quality replacement.
	 * Regardless, show whatever high quality rendering is available as a visual
	 * cue of the rendering progress.
	 */
	void draw_rendering (Cairo.Context cr) {
		if (!rendering_visible) return;
		if (render_job == null) create_image_surface ();
		// Skip low quality rendering if the higher quality is available
		if (!render_job.image_surface_completed || dragging) {
			cr.save ();
			apply_zoom (cr);
			cr.translate (origin_x, origin_y);
			for (int i=0; i < base_pod.size; i++) {
				Pod motif_pod = base_pod.motifs[i];
				var seed = base_pod[i];
				Fractalizer.draw (motif_pod, cr, seed, null, null, false);
			}
			cr.restore ();
		}
		// High quality rendering, as rendered asynchronously
		cr.save ();
		apply_zoom (cr);
		render_job.paint_image_surface (cr, 0, 0);
		cr.restore ();
	}

	/**
	 * Draw the segmentation of each seed.
	 */
	void draw_segmentation (Cairo.Context cr) {
		if (!segmentation_visible) return;
		cr.save ();
		translate_to_pod_origin (cr);
		cr.set_dash ({3, 2}, 0);
		cr.set_line_width (1);
		// Set up a projection square
		var scale = pod_scale * zoom;
		var line = new Line ();
		line.init_full ({0, 0, scale, 0});
		line.base_offset = 0;
		line.num_segment = 0;
		line.width0 = line.width1 = scale;
		var pr = new Projection.by_segment (line, 0, 1, Flip.NONE);
		foreach (var seed in edited_pod) {
			if (seed == selected_seed) cr.set_source_rgb (0, 0, 1);
			else cr.set_source_rgb (1, 0.5, 0);
			seed.draw_segmentation (cr, pr);
		}
		cr.restore ();
	}

	/**
	 * Draw each seed twice. First, using a translucent white/black to fade
	 * the background. Then, using blue. The blue seed will stand out 
	 * from the background.
	 */
	void draw_seeds (Cairo.Context cr) {
		for (int i=0; i < 2; i++) {
			cr.save ();
			translate_to_pod_origin (cr);
			if (i == 0) {
				cr.set_source_rgba (background, background, background, 0.5);
				cr.set_line_width (10);
			}
			else {
				cr.set_source_rgb (0, 0, 1.0);
				cr.set_line_width (1.5);
			}
			foreach (var seed in edited_pod) {
				if (!all_seeds_visible && seed != selected_seed) continue;
				seed.draw_direct (cr, pod_scale * zoom);
			}
			// Show handles of selected seed
			if (selected_seed != null) {
				if (i == 0) {
					cr.set_line_width (10);
					cr.set_source_rgba (background, background, background, 0.5);
				}
				else {
					double[] dashes = {3, 3};
					cr.set_dash (dashes, 0);
					cr.set_line_width (1.0);
					cr.set_source_rgb (1, 0, 0);
				}
				selected_seed.draw_controls (cr, pod_scale * zoom);
			}
			cr.restore ();
		}
	}

	/**
	 * Handler for mouse press.
	 */
	void press (double x, double y, Gdk.ModifierType state) {
		if (zoom_control.press (x, y)) return;
		var snap_on = (state & Gdk.ModifierType.SHIFT_MASK) != 0;
		drawing_area.grab_focus ();
		var p = Point () {x = (image_x (x) - pod_x) / pod_scale, y = (image_y (y) - pod_y) / pod_scale};
		if (selected_seed != null) {
			// Move the seed or its controls
			// Or scroll the widget
			if (!selected_seed.start_drag (p, pod_scale * zoom, true)) {
				selected_seed = null;
			}
		}
		if (selected_seed == null) {
			if (selected_prototype != null) {
				// Apply the tool and set the newly created seed as selected
				// Set selected_seed ONLY AFTER all parameters have been assigned.
				var seed = edited_pod.add_seed_at_point (selected_prototype.get_type (), p);
				seed.copy_parameters (default_seed);
				selected_seed = seed;
				selected_seed.start_drag (p, pod_scale * zoom, true, snap_on, grid_size);
			}
			else {
				selected_seed = seed_at (p);
				// Not dragging any seed, hence scrolling
				if (selected_seed == null) scrolling = true;
			}
		}
		if (scrolling) {
			pressed_x = (int) x;
			pressed_y = (int) y;
			save_centre_x = centre_x;
			save_centre_y = centre_y;
		}
	}

	/**
	 * Handler for mouse release
	 */
	void release (double x, double y) {
		if (zoom_control.release (x, y)) return;
		scrolling = false;
		if (selected_seed != null && dragging) {
			dragging = false;
			edited_pod.signal_changed ();
		}
	}

	/**
	 * Handler for mouse move
	 */
	void move (double x, double y, Gdk.ModifierType state) {
		if (zoom_control.move (x, y)) return;
		var p = Point () {x = (image_x (x) - pod_x) / pod_scale, y = (image_y (y) - pod_y) / pod_scale};
		var snap_on = (state & Gdk.ModifierType.SHIFT_MASK) != 0;
		// Move only. No button is pressed.
		if ((state & Gdk.ModifierType.BUTTON1_MASK) == 0) {
			change_cursor_type (x, y, p);
		}
		// Drag
		else if (scrolling) {
			centre_x = (save_centre_x - (x - pressed_x) / zoom).clamp (0, total_width);
			centre_y = (save_centre_y - (y - pressed_y) / zoom).clamp (0, total_height);
			h_adjustment.value = centre_x;
			v_adjustment.value = centre_y;
			redraw ();
		}
		else if (selected_seed != null) {
			dragging = true;
			selected_seed.drag (p, pod_scale * zoom, snap_on, grid_size, snap_angle);
			redraw ();
		}
	}

	/**
	 * Which seed is pointed by the cursor?
	 */
	Seed? seed_at (Point p) {
		foreach (var seed in edited_pod) {
			if (seed.start_drag (p, pod_scale * zoom, false)) return seed;
		}
		return null;
	}

	/**
	 * x,y : Raw mouse coordinates
	 * p   : Pod coordinates
	 */
	void change_cursor_type (double x, double y, Point p) {
		var window = drawing_area.get_window ();
		var tool_cursor = selected_prototype == null ? Gdk.CursorType.ARROW : Gdk.CursorType.PENCIL;
		var edit_cursor = Gdk.CursorType.HAND1;
		var zoom_cursor = Gdk.CursorType.HAND1;
		if (zoom_control.press (x, y)) {
			zoom_control.release (x, y);
			window.cursor = new Gdk.Cursor (zoom_cursor);
		}
		else if (selected_prototype == null) {
			window.cursor = new Gdk.Cursor (seed_at (p) == null ? tool_cursor : edit_cursor);
		}
		else if (selected_seed != null && selected_seed.start_drag (p, pod_scale * zoom, true)) {
			window.cursor = new Gdk.Cursor (edit_cursor);
		}
		else {
			window.cursor = new Gdk.Cursor (tool_cursor);
		}
	}

	/**
	 * Create text area to view serialized version of edited_pod
	 */
	Gtk.Widget create_text_area () {
		var grid = new Gtk.Grid ();
		grid.orientation = Gtk.Orientation.VERTICAL;
		var text_scroll = new Gtk.ScrolledWindow (null, null);
		text_area = new Gtk.TextView ();
		Util.resolve_accelerator_clashes (text_area);
		grid.add (text_scroll);
		text_scroll.expand = true;
		text_scroll.add (text_area);
		var apply_button = new Gtk.Button.with_label ("Apply changes");
		var button_box = new Gtk.ButtonBox (Gtk.Orientation.HORIZONTAL);
		button_box.add (apply_button);
		grid.add (button_box);
		var auto_change = false;
		edited_pod.signal_changed.connect ((t) => {
			auto_change = true;
			text_area.buffer.text = serialize_pod ();
			auto_change = false;
			apply_button.sensitive = false;
		});
		apply_button.clicked.connect (() => {
			edited_pod.compile (text_area.buffer.text);
			selected_seed = null;
			edited_pod.signal_changed ();
			flip_preview_pod ();
			this.page = 0;
		});
		text_area.buffer.changed.connect (() => {
			if (!auto_change) apply_button.sensitive = true;
		});
		text_area.buffer.text = serialize_pod ();
		apply_button.sensitive = false;
		return grid;
	}

	string serialize_pod () {
		return "# Warning: Auto-generated. User formatting is not preserved.\n" + edited_pod.serialize ();
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
