/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
using Gee;

public class PodEditor : Gtk.Grid {
	Gtk.ApplicationWindow main_window;
	Border border;

	/**
	 * Scratch pad where all the motif editing takes place
	 */
	Pod motif;

	/**
	 * If motif editing is cancelled, the unedited version is
	 * restored from this.
	 */
	Pod save_motif;

	PodDrawing base_drawing;
	PodDrawing motif_drawing;

	Gtk.Notebook main_panel;
	Gtk.Notebook tool_panel;
	MotifChooser motif_chooser;

	/**
	 * Dialogs
	 */
	ImageSizeDialog image_size_dialog = new ImageSizeDialog ();
	SnapDialog snap_dialog = new SnapDialog ();

	/**
	 * File and print dialogs
	 */
	Gtk.PrintSettings print_settings;
	Gtk.FileChooserDialog save_file_chooser;
	Gtk.FileChooserDialog open_file_chooser;
	Gtk.FileChooserDialog export_file_chooser;
	string? current_filename = null;
	string? current_folder = null;

	public PodEditor (Gtk.ApplicationWindow main_window, Border border) {
		this.main_window = main_window;
		create_menu ();
		this.border = border;
		// Initialize our scratch pad
		motif = new Pod ();
		motif.copy (border.default_motif);
		save_motif = new Pod ();

		base_drawing = new PodDrawing.border (border);
		motif_drawing = new PodDrawing.motif (motif);

		Gtk.Grid status_box;
		attach (status_box = create_side_panel (), 0, 1, 1, 1);
		base_drawing.status_box = status_box;
		motif_drawing.status_box = status_box;
		var box = new Gtk.Paned (Gtk.Orientation.VERTICAL);
		attach (box, 1, 1, 1, 1);
		box.add1 (create_gallery ());
		box.add2 (create_main_panel ());
		realize.connect (() => {
			update_status ();
		});
		base_drawing.notify["selected-seed"].connect ((s, p) => { 
			selection_changed ();
			motif_drawing.flip_preview_pod ();
		});
		motif_drawing.notify["selected-seed"].connect ((s, p) => { 
			selection_changed ();
		});
		update_status ();
	}

	Gtk.Widget create_main_panel () {
		main_panel = new Gtk.Notebook ();
		main_panel.show_border = false;
		main_panel.show_tabs = false;
		main_panel.append_page (base_drawing, null);
		main_panel.append_page (create_motif_drawing_page (), null);
		return main_panel;
	}

	Gtk.Widget create_motif_drawing_page () {
		var box = new Gtk.Grid ();
		box.orientation = Gtk.Orientation.VERTICAL;
		box.add (motif_drawing);
		var button_box = new Gtk.ButtonBox (Gtk.Orientation.HORIZONTAL);
		button_box.layout_style = Gtk.ButtonBoxStyle.CENTER;
		button_box.margin_top = 20;
		button_box.margin_bottom = 10;
		box.add (button_box);
		var cancel_button = new Gtk.Button.with_label (_("Cancel"));
		button_box.add (cancel_button);
		cancel_button.clicked.connect (() => {
			// Back to border editing
			main_panel.page = 0;
			tool_panel.page = 0;
			motif.copy (save_motif);
			motif_drawing.flip_preview_pod ();
			update_status ();
		});
		var ok_button = new Gtk.Button.with_mnemonic (_("_Apply Motif"));
		button_box.add (ok_button);
		ok_button.clicked.connect (() => {
			// Back to border editing
			main_panel.page = 0;
			tool_panel.page = 0;
			apply_changes_to_motif ();
			update_status ();
		});
		return box;
	}

	Gtk.Widget create_gallery () {
		motif_chooser = new MotifChooser ();
		motif_chooser.add (border.default_motif);
		foreach (var m in border.motifs) motif_chooser.add (m);
		motif_chooser.changed.connect ((pod) => {
			if (motif.same_as (pod)) return;
			motif.copy (pod);
			motif_drawing.flip_preview_pod ();
			if (main_panel.page == 0) apply_changes_to_motif (); });
		// Layout. Put motif chooser inside a scroller.
		var grid = new Gtk.Grid ();
		var scroller = new Gtk.ScrolledWindow (null, null);
		scroller.hscrollbar_policy = Gtk.PolicyType.NEVER;
		scroller.add (motif_chooser);
		scroller.hexpand = true;
		scroller.vexpand = true;
		grid.attach (scroller, 0, 0, 1, 1);
		return grid;
	}

	Gtk.Grid create_side_panel () {
		var side_panel = new Gtk.Grid ();
		side_panel.orientation = Gtk.Orientation.VERTICAL;
		side_panel.add (new Gtk.Separator (Gtk.Orientation.HORIZONTAL));
		var tool_panel = create_tool_panel ();
		tool_panel.vexpand = true;
		side_panel.add (tool_panel);
		return side_panel;
	}

	struct ActionItem {
		public unowned string name;
		public unowned string? label;
		public string? accel;
	}

	void create_menu () {
		//
		// Note: Any menu items with accelerators clashing with editing keys should be
		// listed in Util.action_names.
		//
		ActionItem[] action_list = {
			ActionItem(){name = "<File",          label = N_("_File")},
			ActionItem()  {name = "New",          label = N_("_New"),               accel = "<control>N"},
			ActionItem()  {name = "Open",         label = N_("_Open..."),           accel = "<control>O"},
			ActionItem()  {name = "-"},
			ActionItem()  {name = "Save",         label = N_("_Save"),              accel = "<control>S"},
			ActionItem()  {name = "SaveAs",       label = N_("Save _As..."),        accel = "<shift><control>S"},
			ActionItem()  {name = "Export",       label = N_("_Export...")},
			ActionItem()  {name = "-"},
			ActionItem()  {name = "Print",        label = N_("_Print..."),          accel = "<control>P"},
			ActionItem()  {name = "-"},
			ActionItem()  {name = "Quit",         label = N_("_Quit"),              accel = "<control>Q"},
			ActionItem()  {name = ">"},
			ActionItem(){name = "<Edit",          label = N_("_Edit")},
			ActionItem()  {name = "Delete",       label = N_("_Delete"),            accel = "Delete"},
			ActionItem()  {name = "-"},
			ActionItem()  {name = "Raise",        label = N_("_Raise"),             accel = "Page_Up"},
			ActionItem()  {name = "Lower",        label = N_("_Lower"),             accel = "Page_Down"},
			ActionItem()  {name = "RaiseTop",     label = N_("Raise To _Top"),      accel = "Home"},
			ActionItem()  {name = "LowerBottom",  label = N_("Lower To _Bottom"),   accel = "End"},
			ActionItem()  {name = "-"},
			ActionItem()  {name = "Previous",     label = N_("Select _Previous"),   accel = "<alt>Left"},
			ActionItem()  {name = "Next",         label = N_("Select _Next"),       accel = "<alt>Right"},
			ActionItem()  {name = "Unselect",     label = N_("Select None"),        accel = "Escape"},
			ActionItem()  {name = "-"},
			ActionItem()  {name = "SnapSetting",  label = N_("_Snap Setting...") },
			ActionItem()  {name = ">"},
			ActionItem(){name = "<Format",        label = N_("F_ormat")},
			ActionItem()  {name = "ImageSize",    label = N_("Image _Size...")},
			ActionItem()  {name = ">"},
			ActionItem(){name = "<View",          label = N_("_View")},
			ActionItem()  {name = "*ViewGrids",   label = N_("Grids"),               accel = "numbersign"},
			ActionItem()  {name = "*ViewSeeds",   label = N_("Seeds") },
			ActionItem()  {name = "+ViewSegment", label = N_("Segmentation")},
			ActionItem()  {name = "*ViewRender",  label = N_("Rendering")},
			ActionItem()  {name = "+ViewDark",    label = N_("Dark Background")},
			ActionItem()  {name = "$-"},
			ActionItem()  {name = "$+ViewText",    label = N_("Text")},
			ActionItem()  {name = ">"},
			ActionItem(){name = "<Help",          label = N_("_Help")},
			ActionItem()  {name = "Contents",     label = N_("_Contents"),           accel = "F1"},
			ActionItem()  {name = "About",        label = N_("_About")},
			ActionItem()  {name = ">"}};
		string menu_xml = "";
		foreach (weak ActionItem action in action_list) {
			var name = action.name;
			var label = action.label == null ? null : _(action.label);
			// Add items whose names start with '$' only in debug mode
			if (name[0] == '$') {
				if (!Util.debug_on ()) continue;
				name = name.substring (1);
			}
			var prefix = name[0];
			if (prefix == '-') {
				menu_xml += "</section><section>";
			}
			else if (prefix == '>') {
				menu_xml += "</section></submenu>";
			}
			else if (prefix == '<') {
				menu_xml += "<submenu id='%s'><attribute name='label'>%s</attribute><section>".printf (name, label);
			}
			else {
				GLib.SimpleAction simple_action;
				if (prefix == '*' || prefix == '+') {
					name = name.substring (1);
					simple_action = new GLib.SimpleAction.stateful (name, null, new Variant.boolean (prefix == '*'));
				}
				else {
					simple_action = new GLib.SimpleAction (name, null);
				}
				simple_action.set_enabled (true);
				simple_action.activate.connect (() => {on_action (name);});
				// TODO: for 'app' scope:
				//  MutiaraApplication.instance.add_action (simple_action);
				main_window.add_action (simple_action);
				var accel_attribute = "";
				if (action.accel != null) {
					accel_attribute = "<attribute name='accel'>%s</attribute>" .printf (Markup.escape_text (action.accel));
				}
				// TODO: all the actions currently have 'win' scope.
				// Some should have 'app' scope
				menu_xml += ("<section><item>" +
				             "<attribute name='label'>%s</attribute>" +
				             "<attribute name='action'>win.%s</attribute>" +
							 "%s" +
				             "</item></section>") .printf (label, name, accel_attribute);
			}
		}

		Gtk.Builder builder = new Gtk.Builder ();
		try {
			builder.add_from_string (
				"<interface>" +
				"  <menu id='MenuBar'>" +
				"   " + menu_xml +
				"  </menu>" +
				"</interface>",
				-1  // String length. -1 means null-terminated. This has no business in vala!!
			);
		}
		catch (Error e) {
			stderr.printf ("Fail to build menu: %s\n", e.message);
			assert_not_reached ();
		}

		MutiaraApplication.instance.set_menubar (builder.get_object ("MenuBar") as GLib.MenuModel);
	}

	/**
	 * Dispatch actions to their respective handlers
	 */
	void on_action (string action_string) {
		var pod_drawing = main_panel.page == 0 ? base_drawing : motif_drawing;
		var action = main_window.lookup_action (action_string);
		bool ticked = false;
		if (action.state_type != null) {
			ticked = ! action.state.get_boolean ();
			action.change_state (ticked);
		}
		switch (action_string) {
		case "New" :
			perform_action_new ();
			break;
		case "Open" :
			perform_action_open ();
			break;
		case "Save" :
			perform_action_save ();
			break;
		case "SaveAs" :
			perform_action_saveas ();
			break;
		case "Export" :
			perform_action_export ();
			break;
		case "Print" :
			perform_action_print ();
			break;
		case "Quit" :
			main_window.hide ();
			main_window.destroy ();
			break;

		case "Delete" :
			pod_drawing.delete_selection ();
			break;
		case "Previous" :
			pod_drawing.select_previous ();
			break;
		case "Next" :
			pod_drawing.select_next ();
			break;
		case "Unselect" :
			pod_drawing.selected_seed = null;
			break;
		case "Lower" :
			pod_drawing.move_selection_down ();
			selection_changed ();
			break;
		case "Raise" :
			pod_drawing.move_selection_up ();
			selection_changed ();
			break;
		case "LowerBottom" :
			pod_drawing.move_selection_to_bottom ();
			selection_changed ();
			break;
		case "RaiseTop" :
			pod_drawing.move_selection_to_top ();
			selection_changed ();
			break;
		case "SnapSetting" :
			snap_dialog.run (main_window, pod_drawing);
			break;
		case "ImageSize" :
			image_size_dialog.run (main_window, base_drawing);
			break;

		case "ViewText" :
			if (!ticked) {
				base_drawing.page = 0;
				motif_drawing.page = 0;
				base_drawing.show_tabs = false;
				motif_drawing.show_tabs = false;
			}
			else {
				base_drawing.show_tabs = true;
				motif_drawing.show_tabs = true;
			}
			break;
		case "ViewGrids" :
			base_drawing.grid_visible = ticked;
			motif_drawing.grid_visible = ticked;
			break;
		case "ViewSeeds" :
			base_drawing.all_seeds_visible = ticked;
			motif_drawing.all_seeds_visible = ticked;
			break;
		case "ViewSegment" :
			base_drawing.segmentation_visible = ticked;
			motif_drawing.segmentation_visible = ticked;
			break;
		case "ViewRender" :
			base_drawing.rendering_visible = ticked;
			motif_drawing.rendering_visible = ticked;
			break;
		case "ViewDark" :
			base_drawing.dark_background = ticked;
			motif_drawing.dark_background = ticked;
			motif_chooser.dark_background = ticked;
			break;
		case "Contents" :
			try {
				Gtk.show_uri (main_window.screen, "ghelp:mutiara",  Gtk.get_current_event_time ());
			}
			catch (Error e) {
				stderr.printf("Error: %s \n", e.message);
			}
			break;
		case "About" :
			show_about_dialog ();
			break;
		}
	}

	void perform_action_new () {
		var pod_drawing = main_panel.page == 0 ? base_drawing : motif_drawing;
		pod_drawing.edited_pod.clear ();
		pod_drawing.edited_pod.signal_changed ();
		if (pod_drawing == base_drawing) {
			on_action ("ImageSize");
			current_filename = null;
		}
		update_status ();
	}

	void perform_action_save () {
		if (current_filename == null) perform_action_saveas ();
		else save_to_file (current_filename);
	}

	void perform_action_saveas () {
		if (save_file_chooser == null) {
			save_file_chooser = new Gtk.FileChooserDialog (_("Save"), main_window, Gtk.FileChooserAction.SAVE,
				_("Cancel"), Gtk.ResponseType.CANCEL, _("Save"), Gtk.ResponseType.OK);
			save_file_chooser.do_overwrite_confirmation = true;
			add_file_filter_to_dialog (save_file_chooser);
		}
		if (current_folder != null) save_file_chooser.set_current_folder (current_folder);
		save_file_chooser.set_current_name (current_filename != null ?
				Path.get_basename (current_filename) :
				"new drawing.mutiara");
		var response = save_file_chooser.run ();
		save_file_chooser.hide ();
		if (response != Gtk.ResponseType.OK) return;
		current_folder = save_file_chooser.get_current_folder ();
		var file = save_file_chooser.get_file ();
		var filename = file.get_path ();
		//if (!filename.has_suffix (".mutiara")) filename += ".mutiara"; // Doesn't work with overwrite_confirmation
		save_to_file (filename);
	}

	void save_to_file (string filename) {
		string contents = "#mutiara 1.0\n";
		contents += "Size %d %d\n".printf (base_drawing.pod_width, base_drawing.pod_height);
		contents += "\nDrawing\n";
		contents += border.serialize ();
		try {
			FileUtils.set_contents (filename, contents);
			current_filename = filename;
			update_status ();
		}
		catch (Error e) {
			// TODO Show error in dialog
			stderr.printf ("File error: %s\n", e.message);
		}
	}

	void perform_action_open () {
		if (open_file_chooser == null) {
			open_file_chooser = new Gtk.FileChooserDialog (_("Open"), main_window, Gtk.FileChooserAction.OPEN,
				_("Cancel"), Gtk.ResponseType.CANCEL, _("Open"), Gtk.ResponseType.OK);
			add_file_filter_to_dialog (open_file_chooser);
		}
		if (current_folder != null) open_file_chooser.set_current_folder (current_folder);
		var response = open_file_chooser.run ();
		open_file_chooser.hide ();
		if (response != Gtk.ResponseType.OK) return;
		current_folder = open_file_chooser.get_current_folder ();
		var file = open_file_chooser.get_file ();
		var filename = file.get_path ();
		read_from_file (filename);
	}

	void read_from_file (string filename) {
		string contents;
		try {
			FileUtils.get_contents (filename, out contents);
		}
		catch (Error e) {
			// TODO Show error in dialog
			stderr.printf ("File error: %s\n", e.message);
			return;
		}
		// ss = { "page setup", "drawing" }
		var ss = contents.split ("\nDrawing\n");
		if (ss.length < 2) return;
		// Get the set up
		foreach (var line in ss[0].split ("\n")) {
			// line = {"command", "args"}
			var cmdarg = line.strip ().split (" ", 2);
			if (cmdarg.length == 0) continue;
			double[] arg = {};
			if (cmdarg.length == 2) {
				foreach (var token in Regex.split_simple ("[\\(|\\)|\\,|\\s]++", cmdarg[1])) {
					if (token.length == 0) continue;
					arg += double.parse (token);
				}
				// Other arguments default to 0.0
				for (int i=0; i < 10; i++) arg += 0.0;
			}
			switch (cmdarg[0]) {
				case "Size" :
					base_drawing.resize (int.max (10, (int) arg[0]), int.max (10, (int) arg[1]));
					break;
			}
		}
		border.compile (ss[1]);
		border.signal_changed ();
		// Puts the motifs used in the file inside the gallery
		foreach (var m in border.motifs) motif_chooser.add (m);
		current_filename = filename;
		update_status ();
	}

	void add_file_filter_to_dialog (Gtk.FileChooserDialog file_chooser) {
		var filter = new Gtk.FileFilter ();
		filter.add_pattern ("*");
		filter.set_filter_name (_("All files"));
		file_chooser.add_filter (filter);
		filter = new Gtk.FileFilter ();
		filter.add_pattern ("*.mutiara");
		filter.set_filter_name (_("Mutiara files"));
		file_chooser.add_filter (filter);
		file_chooser.set_filter (filter);
	}

	void perform_action_export () {
		if (export_file_chooser == null) {
			export_file_chooser = new Gtk.FileChooserDialog (_("Export"), main_window, Gtk.FileChooserAction.SAVE,
				_("Cancel"), Gtk.ResponseType.CANCEL, _("Save"), Gtk.ResponseType.OK);
			export_file_chooser.do_overwrite_confirmation = true;
			var filter = new Gtk.FileFilter ();
			filter.add_mime_type ("image/png");
			filter.set_filter_name (_("PNG - Portable Network Graphic (.png)"));
			export_file_chooser.add_filter (filter);
			export_file_chooser.set_filter (filter);
		}
		if (current_folder != null) export_file_chooser.set_current_folder (current_folder);
		export_file_chooser.set_current_name (current_filename != null ?
				Path.get_basename (current_filename) + ".png" :
				"new drawing.png");
		var response = export_file_chooser.run ();
		export_file_chooser.hide ();
		if (response != Gtk.ResponseType.OK) return;
		current_folder = export_file_chooser.get_current_folder ();
		var file = export_file_chooser.get_file ();
		var filename = file.get_path ();
		//if (!filename.has_suffix (".png")) filename += ".png"; // Doesn't work with overwrite_confirmation
		base_drawing.save_image_surface.begin (filename);
	}

	void perform_action_print () {
		if (print_settings == null) {
			print_settings = new Gtk.PrintSettings ();
			//print_settings.set_resolution (300); // has no effect on Cairo resolution
		}
		var print_operation = new Gtk.PrintOperation ();
		print_operation.print_settings = print_settings;
		print_operation.draw_page.connect ((ctx, pg) => { base_drawing.print.begin (print_operation, ctx); });
		print_operation.n_pages = 1;
		//print_operation.allow_async = true;
		try {
			print_operation.run (Gtk.PrintOperationAction.PRINT_DIALOG, main_window);
			// TODO Is the following line really appropriate?
			print_settings = print_operation.print_settings;
		}
		catch (Error e) {
			// TODO Show error in dialog
			stderr.printf ("File error: %s\n", e.message);
		}
	}

	/**
	 * To be called whenever the user select/deselect
	 */
	void selection_changed () {
		var pod_drawing = main_panel.page == 0 ? base_drawing : motif_drawing;
		var pod = pod_drawing.edited_pod;
		var seed = pod_drawing.selected_seed;
		var n = pod.index_of (seed);
		var size = pod.size;
		(main_window. lookup_action ("Delete") as GLib.SimpleAction). set_enabled (seed != null);
		(main_window. lookup_action ("Previous") as GLib.SimpleAction). set_enabled (size > 0 && (seed == null || n > 0));
		(main_window. lookup_action ("Next") as GLib.SimpleAction). set_enabled (size > 0 && (seed == null || n < size - 1));
		(main_window. lookup_action ("Lower") as GLib.SimpleAction). set_enabled (seed != null && n > 0);
		(main_window. lookup_action ("Raise") as GLib.SimpleAction). set_enabled (seed != null && n != size - 1);
		(main_window. lookup_action ("LowerBottom") as GLib.SimpleAction). set_enabled (seed != null && n > 0);
		(main_window. lookup_action ("RaiseTop") as GLib.SimpleAction). set_enabled (seed != null && n != size - 1);
	}

	Gtk.Widget create_tool_panel () {
		tool_panel = new Gtk.Notebook ();
		tool_panel.show_border = false;
		tool_panel.show_tabs = false;
		PodPanel base_panel = new PodPanel (main_window, base_drawing, motif);
		tool_panel.append_page (base_panel, null);
		tool_panel.append_page (new PodPanel (main_window, motif_drawing, null), null);
		base_panel.motif_button.clicked.connect (() => {
			// Show motif editing page
			save_motif.copy (motif);
			tool_panel.page = 1;
			main_panel.page = 1;
			update_status ();
		});
		return tool_panel;
	}

	void update_status () {
		selection_changed ();
		main_window.title = (current_filename ?? _("New Mutiara Drawing"));
		base_drawing.hidden = main_panel.page == 1;
		motif_drawing.hidden = main_panel.page == 0;
	}

	/**
	 * When view is switched from motif editing to border editing, changes to
	 * the motif should be applied to the context of the border editing.
	 */
	void apply_changes_to_motif () {
		if (base_drawing.selected_seed == null) {
			// Apply changes to default motif
			base_drawing.default_seed.flip_even = motif.flip_even;
			base_drawing.default_seed.flip_odd = motif.flip_odd;
			base_drawing.default_seed.tweak = motif.tweak;
			border.default_motif = new Pod ();
			border.default_motif.copy (motif);
			base_drawing.selected_seed = null;  // notify changes
		}
		else {
			// Apply changes to seed's motif
			var n = border.index_of (base_drawing.selected_seed);
			if (! border.motifs[n].same_as (motif)) {
				base_drawing.selected_seed.flip_even = motif.flip_even;
				base_drawing.selected_seed.flip_odd = motif.flip_odd;
				base_drawing.selected_seed.tweak = motif.tweak;
				border.motifs[n] = new Pod ();
				border.motifs[n].copy (motif);
				border.signal_changed ();
				base_drawing.selected_seed = base_drawing.selected_seed;  // notify changes
			}
		}
		motif_chooser.add (motif);
	}

	void show_about_dialog () {
		Gdk.Pixbuf logo = null;
		try {
			logo = new Gdk.Pixbuf.from_file (Config.PIXMAPSDIR + "/mutiara-logo.png");
		}
		catch (Error e) {
			stderr.printf ("Cannot load logo: %s\n", e.message);
		}
		Gtk.show_about_dialog (main_window,
			"program-name",  _(Config.PACKAGE_NAME),
			"version", Config.PACKAGE_VERSION,
			"comments", _("Design motifs the fun way"),
			"copyright", "Copyright (C) 2011 Nor Jaidi Tuah <norjaidi.tuah@ubd.edu.bn>",
			"logo", logo,
			"authors", new string[]{
				"Nor Jaidi Tuah <norjaidi.tuah@ubd.edu.bn>" },
			"artists", new string[]{"Nor Jaidi Tuah <norjaidi.tuah@ubd.edu.bn>"},
			"translator_credits", _("translator_credits"),
            "license-type", Gtk.License.GPL_3_0,
            "website", "http://mutiara.sourceforge.net/",
            "website-label", "mutiara.sourceforge.net"
		);
	}

}
// vim: tabstop=4: textwidth=100: autoindent:
