/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
using Gee;

public class PodPanel : Gtk.Grid {
	Gtk.Window parent_window;
	PodDrawing pod_drawing;

	/**
	 * A link to the scratch pad created in PodEditor
	 */
	Pod motif;

	/**
	 * Either editing motif or editing base.
	 */
	bool editing_motif;

	/**
	 * Drawing and editing tools
	 */
	Seed[] prototypes = {new Line (), new Arc (), new Curve ()};

	/**
	 * Widget to enter seed parameters.
	 */
	bool ignore_changes;
	Gtk.SpinButton num_segment_entry;
	Gtk.SpinButton width0_entry;
	Gtk.SpinButton width1_entry;
	Gtk.ComboBox base_offset_entry;
	Gtk.ComboBox role_entry;
	TweakWidget tweak_entry;
	Gtk.SpinButton level_of_detail_entry;
	Gtk.ColorButton color_entry;
	Gtk.SpinButton color_mix_entry;
	Gtk.ComboBox render_style_entry;
	public MotifButton motif_button;
	delegate void EntryMethod ();

	/**
	 * Widget for parameters applied to the entire motif
	 */
	TweakWidget entire_tweak_entry;

	public PodPanel (Gtk.Window parent_window, PodDrawing pod_drawing, Pod? motif) {
		this.parent_window = parent_window;
		this.pod_drawing = pod_drawing;
		this.motif = motif;
		editing_motif = !(pod_drawing.edited_pod is Border);

		this.orientation = Gtk.Orientation.VERTICAL;
		add (create_edit_toolbar ());
		add (new Gtk.Separator (Gtk.Orientation.HORIZONTAL));
		add (create_widget_for_seed_options ());

		realize.connect (update_parameter_area);
		pod_drawing.notify["selected-seed"].connect ((s, p) => { 
			update_parameter_area (); 
		});
	}

	/**
	 * Widget to display editing tools.
	 */
	Gtk.Widget create_edit_toolbar () {
		var toolbar = new Gtk.Grid ();
		toolbar.orientation = Gtk.Orientation.HORIZONTAL;
		// Selection of drawing tools
		var arrow_button = create_null_tool ();
		toolbar.add (arrow_button);
		foreach (var seed in prototypes) {
			toolbar.add (create_drawing_tool (seed, arrow_button));
		}
		return toolbar;
	}

	/**
	 * Create a user button to deselect all drawing tools
	 */
	Gtk.RadioButton create_null_tool () {
		var button = new Gtk.RadioButton (null);
		button.can_focus = false;
		button.relief = Gtk.ReliefStyle.NONE;
		button.set_mode (false);
		button.border_width = 0;
		button.clicked.connect ((s) => {pod_drawing.selected_prototype = null;});
		var image = new Gtk.DrawingArea ();
		var size = 15;
		image.set_size_request (size, size);
		image.set_has_window (false);
		image.draw.connect ((cr) => {
			cr.set_line_width (2.5);
			cr.move_to (0.2 * size, 0.8 * size);
			cr.line_to (0.2 * size, 0.1 * size);
			cr.line_to (0.8 * size, 0.4 * size);
			cr.move_to (0.2 * size, 0.1 * size);
			cr.line_to (0.75 * size, 0.9 * size);
			cr.stroke ();
			return true;});
		button.image = image;
		return button;
	}

	/**
	 * Create a user button for drawing the given seed
	 */
	Gtk.RadioButton create_drawing_tool (Seed prototype, Gtk.RadioButton group) {
		var button = new Gtk.RadioButton.from_widget (group);
		button.can_focus = false;
		button.relief = Gtk.ReliefStyle.NONE;
		button.set_mode (false);
		button.border_width = 0;
		button.clicked.connect ((s) => {pod_drawing.selected_prototype = prototype;});
		var image = new Gtk.DrawingArea ();
		var size = 15;
		image.set_size_request (size, size);
		image.set_has_window (false);
		image.draw.connect ((cr) => {
			cr.set_line_width (1.0);
			prototype.draw_icon (cr, size, size);
			return true;});
		button.image = image;
		return button;
	}

	/**
	 * Utility methods to create various widgets for parameter entries
	 */
	Gtk.ComboBox create_combobox (string[] options) {
		var widget = new Gtk.ComboBoxText ();
		foreach (var option in options) widget.append_text (option);
		return widget;
	}

	Gtk.ComboBox create_base_offset_entry () {
		return create_combobox ({_("Top"), _("Middle"), _("Bottom")});
	}

	Gtk.ComboBox create_role_entry () {
		return create_combobox ({_("Dual"), _("Motif"), _("Base")});
	}

	/**
	 * Attach the widget visually to the table
	 */
	void attach_seed_widget (Gtk.Grid table, int y, string? label_text, Gtk.Widget? widget) {
		Util.attach (table, y, label_text, widget);
		if (widget is Gtk.SpinButton) (widget as Gtk.SpinButton).value_changed.connect (read_parameter_area);
		if (widget is Gtk.ColorButton) (widget as Gtk.ColorButton).color_set.connect (read_parameter_area);
		if (widget is Gtk.ComboBox) (widget as Gtk.ComboBox).changed.connect (read_parameter_area);
		if (widget is TweakWidget) (widget as TweakWidget).tweaked.connect (read_parameter_area);
	}

	/**
	 * Read entries in the parameter area onto the selected seed.
	 */
	void read_parameter_area () {
		if (ignore_changes) return;
		var seed = pod_drawing.selected_seed ?? pod_drawing.default_seed;
		seed.num_segment = (int) (num_segment_entry.value);
		seed.width0 = width0_entry.value;
		seed.width1 = editing_motif ? width1_entry.value : width0_entry.value;
		seed.base_offset = base_offset_entry.active * 0.5;
		if (role_entry != null) seed.role = (Role) (role_entry.active);
		if (level_of_detail_entry != null) seed.level_of_detail = (int) (level_of_detail_entry.value);
		if (tweak_entry != null) tweak_entry.get_tweaks (ref seed.tweak, out seed.flip_even, out seed.flip_odd);

		if (color_entry != null) {
			// Gdk to Cairo color
			var rgba = color_entry.rgba;
			seed.color = RGBA (rgba.red, rgba.green, rgba.blue, rgba.alpha);
		}
		if (color_mix_entry != null) seed.color_mix = color_mix_entry.value;
		if (render_style_entry != null) seed.render_style = (RenderStyle) (render_style_entry.active);

		if (entire_tweak_entry != null) {
			entire_tweak_entry.get_tweaks (ref pod_drawing.edited_pod.tweak,
			                              out pod_drawing.edited_pod.flip_even,
			                              out pod_drawing.edited_pod.flip_odd);
			pod_drawing.flip_preview_pod ();
		}
		// Update drawing unless no seed is selected
		if (seed != pod_drawing.default_seed) pod_drawing.edited_pod.signal_changed ();
	}

	/**
	 * Update parameter entries based on the currently selected seed.
	 */
	void update_parameter_area () {
		Seed seed;
		if (pod_drawing.selected_seed == null) {
			seed = pod_drawing.default_seed;
			// Configuration for newly created shape
			if (motif_button != null) {
				motif.copy (pod_drawing.base_pod.default_motif);
			}
		}
		else {
			seed = pod_drawing.selected_seed;
			var n = pod_drawing.edited_pod.index_of (seed);
			if (motif_button != null) {
				motif.copy (pod_drawing.base_pod.motifs[n]);
			}
		}
		ignore_changes = true;

		num_segment_entry.value = seed.num_segment;
		width0_entry.value = seed.width0;
		if (editing_motif) width1_entry.value = seed.width1;
		base_offset_entry.active = int.max (0, int.min (2, (int) (seed.base_offset * 2)));
		if (role_entry != null) role_entry.active = seed.role;
		if (level_of_detail_entry != null) level_of_detail_entry.value = seed.level_of_detail;
		if (tweak_entry != null) tweak_entry.set_tweaks (seed.tweak, seed.flip_even, seed.flip_odd);

		if (color_entry != null) {
			// Cairo to Gdk color
			var rgba = Gdk.RGBA() {
			              red = seed.color.r,
			              green = seed.color.g,
			              blue = seed.color.b,
			              alpha = seed.color.a};
			color_entry.rgba = rgba;
		}
		if (color_mix_entry != null) color_mix_entry.value = seed.color_mix;
		if (render_style_entry != null) render_style_entry.active = seed.render_style;

		if (entire_tweak_entry != null) entire_tweak_entry.set_tweaks (
		                                     pod_drawing.edited_pod.tweak,
		                                     pod_drawing.edited_pod.flip_even,
		                                     pod_drawing.edited_pod.flip_odd);

		ignore_changes = false;
	}

	Gtk.Widget create_widget_for_seed_options () {
		var grid = new Gtk.Grid ();
		grid.set_column_spacing (5);
		grid.set_row_spacing (3);
		grid.margin_left = 5;
		grid.margin_right = 5;
		int y = 0;

		// Segment parameters
		grid.attach (Util.title (_("Segmentation")), 0, y++, 2, 1);
		attach_seed_widget (grid, y++, _("Base:"), base_offset_entry = create_base_offset_entry ());
		if (editing_motif) {
			attach_seed_widget (grid, y++, _("Split:"), num_segment_entry = Util.make_spin_button (1, 100, 1));
			attach_seed_widget (grid, y++, _("Width 1:"), width0_entry = Util.make_spin_button (0, 1, 0.05));
			attach_seed_widget (grid, y++, _("Width 2:"), width1_entry = Util.make_spin_button (0, 1, 0.05));
		}
		else {
			attach_seed_widget (grid, y++, _("Split:"), num_segment_entry = Util.make_spin_button (1, 1000, 1));
			attach_seed_widget (grid, y++, _("Width:"), width0_entry = Util.make_spin_button (0, 10000, 50));
		}

		// Replication parameters
		grid.attach (Util.title (_("Replication")), 0, y++, 2, 1);
		if (editing_motif) {
			attach_seed_widget (grid, y++, _("Role:"), role_entry = create_role_entry ());
			attach_seed_widget (grid, y++, _("Tweak:"), tweak_entry = new TweakWidget (parent_window, true));
		}
		else {
			attach_seed_widget (grid, y++, _("Motif:"), motif_button = new MotifButton (motif));
			attach_seed_widget (grid, y++, _("Flip:"), tweak_entry = new TweakWidget (parent_window, false));
		}
		attach_seed_widget (grid, y++, _("Detail:"), level_of_detail_entry = Util.make_spin_button (0, 100, 1));

		// Rendering parameters
		if (editing_motif) {
			grid.attach (Util.title (_("Rendering")), 0, y++, 2, 1);
			attach_seed_widget (grid, y++, _("Colour:"), color_entry = new Gtk.ColorButton ());
			attach_seed_widget (grid, y++, _("Mix:"),    color_mix_entry = Util.make_spin_button (0, 1, 0.05));
			color_entry.use_alpha = true;
			attach_seed_widget (grid, y++, _("Style:"),
				render_style_entry = create_combobox ({_("Thin"), _("Thick"), _("Hollow"), _("Filled"), _("Closed"), _("None")}));
		}

		// Default parameters for the entire motif
		if (editing_motif) {
			grid.attach (Util.title (_("Entire motif")), 0, y++, 2, 1);
			attach_seed_widget (grid, y++, _("Tweak:"), entire_tweak_entry = new TweakWidget (parent_window, true));
		}
		return grid;
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
