/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
public struct Point {
	public double x;
	public double y;

	public static double hypot (Point p0, Point p1) {
		return Math.hypot (p1.y - p0.y, p1.x - p0.x);
	}

	public Point rotate (double angle) {
		double sin;
		double cos;
		Math.sincos (angle, out sin, out cos);
		return Point () {
			x = x * cos - y * sin,
			y = x * sin + y * cos};
	}

	public Point scale (double sx, double sy) {
		return Point () {x = x * sx, y = y * sy};
	}

	public Point move (Point new_origin) {
		return Point () {x = x + new_origin.x, y = y + new_origin.y};
	}

	public Point move_xy (double x, double y) {
		return Point () {x = this.x + x, y = this.y + y};
	}
}

public enum Role {
	/**
	 * Seed replicated starting at depth 0
	 */
	DUAL = 0,
	/**
	 * Seed replicated starting at depth 1
	 */
	MOTIF = 1,
	/**
	 * Seed replicated at depth 0, but disappears at depth 1
	 */
	BASE = 2,
}

public enum RenderStyle {
	THIN = 0,
	THICK = 1,
	HOLLOW = 2,
	FILLED = 3,
	CLOSED = 4,
	NONE = 5,
}

public class Projection {
	weak Seed? seed;
	Projection* parent;

	// Precalculated affine projection. Precalculation is done in constructor.
	//  [m0 m1 0]
	//  [m2 m3 0]
	//  [m4 m5 1]
	double m[6];

	/**
	 * The projection represented by segment t0-t1 of seed.
	 *
	 * Project using the following reference points:
	 *   (0,base_offset) -> point on projection.seed at t0
	 *   (1,base_offset) -> point on projection.seed at t1
	 * For other points:
	 *   (x,base_offset) -> point on projection.seed at t = t0 + x/(t1-t0)
	 *   (x,y)   -> point at a distance of base_offset-y from (x,base_offset)
	 */
	public Projection.by_segment (Seed seed, double t0, double t1, Flip flip) {
		this.seed = seed;
		// Identity
		m[0] = 1; m[3] = 1;
		// Rotate and scale about (0.5, 0.5), i.e., the centre of the segment
		//   translate -0.5, -0.5
		m[4] = -0.5; m[5] = -0.5;
		//   scale according to seed.tweak
		add (seed.tweak.scale_x, 0, 0, seed.tweak.scale_y);
		//   rotate according to seed.tweak
		var angle = seed.tweak.rotate * Math.PI / 180;
		double sin, cos;
		Math.sincos (angle, out sin, out cos);
		add (cos, sin, -sin, cos);
		//   translate according to seed.tweak
		// Translate 0.5, 0.5 to undo the very first translation
		m[4] += seed.tweak.move_x + 0.5;
		m[5] += seed.tweak.move_y + 0.5;
		// Flip
		if (flip.x ()) {    // x = 1 - x
			add (-1, 0, 0, 1);
			m[4] += 1;
		}
		if (flip.y ()) {    // y = 1 - y
			add (1, 0, 0, -1);
			m[5] += 1;
		}
		// Project onto the segment
		//  x = t0 + x * (t1 - t0)
		//  y = seed.base_offset - y
		add (t1 - t0, 0, 0, -1);
		m[4] += t0;
		m[5] += seed.base_offset;
		
		// Merge with the transformation otherwise done by Line.distant_point
		if (seed.is_affine) {
			var line = seed as Line;
			var dx = line.p1.x - line.p0.x;
			var dy = line.p1.y - line.p0.y;
			var w_cos = line.width0 * line.cos_perpendicular_angle;
			var w_sin = line.width0 * line.sin_perpendicular_angle;
			var m0 = dx * m[0] + w_cos * m[1];
			var m2 = dx * m[2] + w_cos * m[3];
			var m4 = dx * m[4] + w_cos * m[5] + line.p0.x;
			var m1 = dy * m[0] + w_sin * m[1];
			var m3 = dy * m[2] + w_sin * m[3];
			var m5 = dy * m[4] + w_sin * m[5] + line.p0.y;
			m[0] = m0; m[1] = m1; m[2] = m2; m[3] = m3; m[4] = m4; m[5] = m5;
			this.seed = null;   // No need to call seed.distant_point (see Projection.map)
		}
	}

	/**
	 * Copy
	 */
	public void copy_projection (Projection* source) {
		seed = source->seed;
		m = source->m;
		parent = source->parent;
	}
	
	/**
	 * Copy source and link it to a parent.
	 */
	public void compose (Projection* parent, Projection* source) {
		seed = source->seed;
		m = source->m;
		if (seed == null && parent != null) {
			// Merge with parent projection
			add (parent->m[0], parent->m[1], parent->m[2], parent->m[3], parent->m[4], parent->m[5]);
			seed = parent->seed;
			this.parent = parent->parent;
		}
		else {
			this.parent = parent;
		}
	}

	/**
	 * [1 0 0]
	 * [0 1 0]
	 * [0 0 1]
	 */
	public void compose_unit_matrix () {
		m[0] = 1; m[1] = 0; m[2] = 0; m[3] = 1; m[4] = 0; m[5] = 0;
		parent = null;
		seed = null;
	}
	
	/**
	 * Return p * abcedf * seed
	 */
	public static Point map (Projection* pr, Point p) {
		while (pr != null) {
			// Apply m (linear), and then project the segment itself (possibly non-linear)
			var x = pr->m[0] * p.x  +  pr->m[2] * p.y  +  pr->m[4];
			var y = pr->m[1] * p.x  +  pr->m[3] * p.y  +  pr->m[5];
			if (pr->seed == null) {
				p.x = x;
				p.y = y;
			}
			else {
				p = pr->seed.distant_point (x, y);
			}
			pr = pr->parent;
		}
		return p;
	}

	/**
	 * abcdef = abcdef * pqrstu
	 */
	void add (double p, double q, double r, double s, double t = 0, double u = 0) {
		double z[6];
		z[0] = m[0] * p  +  m[1] * r;
		z[1] = m[0] * q  +  m[1] * s;
		z[2] = m[2] * p  +  m[3] * r;
		z[3] = m[2] * q  +  m[3] * s;
		z[4] = m[4] * p  +  m[5] * r + t;
		z[5] = m[4] * q  +  m[5] * s + u;
		m = z;
	}
}

public struct RGBA {
	public double r;
	public double g;
	public double b;
	public double a;

	public RGBA (double r, double g, double b, double a) {
		this.r = r;
		this.g = g;
		this.b = b;
		this.a = a;
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
