/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
using Gee;

public interface IRender : Object {
	public abstract void render (double fraction);
}

public class RenderJob : Object, IRender {
	/**
	 * (Mutexed)
	 */
	Cairo.ImageSurface image_surface;
	public bool image_surface_completed;
	/**
	 * Private copy, not merely a reference, of the border to render.
	 */
	Border border;
	int origin_x;
	int origin_y;
	int width;
	int height;
	/**
	 * Number of requests for this job.
	 * (Mutexed)
	 */
	int request_count;
	/**
	 * Created upon the first request for this job to start its rendering work.
	 * cancelled when request_count is 0.
	 */
	Probe probe = null;
	/**
	 * List of listener to be notified periodically.
	 * (Mutexed)
	 */
	ArrayList<IRender> listeners = null;
	bool shared;

	/**
	 * Thread pool. Jobs pushed to the thread pool is referenced in pushed_jobs
	 * to prevent it from being destroyed while a thread is still accessing it.
	 */
	Timer timer = new Timer ();
	static ThreadPool<RenderJob> thread_pool = null;
	static LinkedList<RenderJob> pushed_jobs = null;

	/**
	 * List of shared jobs. (i.e. shared between display, saving and printing)
	 * (Mutexed)
	 */
	static ArrayList<RenderJob> jobs = null;
	static bool cleaning = false;

	/**
	 * List of unshared named jobs
	 */
	static HashMap<string,RenderJob> named_jobs = null;

	/**
	 * Listeners with pending messages to be sent during idle.
	 * The list is consolidated so that no listener appears twice.
	 */
	static ArrayList<IRender> pending_listeners;
	static ArrayList<double?> messages;

	/**
	 * If the job submitted has the same key as a previous one, continue that previous one
	 * if the border being rendered is the same (this may happen due to spurious change signals
	 * being sent)
	 */
	public static RenderJob create_named (string key, Border border, int origin_x, int origin_y, int width, int height, IRender? listener) {
		init_thread_pool ();
		lock (named_jobs) {
			if (named_jobs == null) named_jobs = new HashMap<string,RenderJob> ();
			if (named_jobs.has_key (key)) {
				var job = named_jobs[key];
				if (job.border.same_as (border)) return job;
				named_jobs.unset (key);
				job.cancel ();
			}
			var job = create (border, origin_x, origin_y, width, height, listener);
			job.shared = false;
			named_jobs[key] = job;
			job.probe = new Probe ();
			queue (job);
			return job;
		}
	}

	/**
	 * If the job submitted has the same spec as the last one, simply return the last one.
	 */
	public static RenderJob create_shared (Border border, int origin_x, int origin_y, int width, int height, IRender? listener) {
		init_thread_pool ();
		lock (jobs) {
			if (jobs == null) jobs = new ArrayList<RenderJob> ();
			remove_finished_jobs ();
			// most recently created job
			var mrj = jobs.size == 0 ? null : jobs.last ();
			if (mrj == null || mrj.origin_x != origin_x || mrj.origin_y != origin_y ||
				mrj.width < width || mrj.height < height || !mrj.border.same_as (border)) {
				var job = create (border, origin_x, origin_y, width, height, listener);
				job.shared = true;
				jobs.add (job);
				remove_finished_jobs ();   // sic. called twice
				job.probe = new Probe ();
				queue (job);
				return job;
			}
			else {
				lock (mrj.listeners) {
					mrj.request_count ++;
					if (listener != null) mrj.listeners.add (listener);
				}
				return mrj;
			}
		}
	}

	static RenderJob create (Border border, int origin_x, int origin_y, int width, int height, IRender? listener) {
		var job = new RenderJob ();
		job.border = new Border (border.default_motif);
		job.border.copy (border);
		job.origin_x = origin_x;
		job.origin_y = origin_y;
		job.width = width;
		job.height = height;
		job.image_surface = new Cairo.ImageSurface (Cairo.Format.ARGB32, job.width, job.height);
		job.request_count = 1;
		job.listeners = new ArrayList<IRender> ();
		if (listener != null) job.listeners.add (listener);
		return job;
	}

	/**
	 * A job is removable if its request_count is 0 and it is not the last one created.
	 */
	static void remove_finished_jobs () {
		lock (cleaning) {
			if (cleaning) return;
			cleaning = true;
		}
		lock (jobs) {
			cleaning = true;
			for (var i = jobs.size - 2; i >= 0; i--) {
				var job = jobs[i];
				if (job.request_count <= 0) job.probe.cancel ();
				if (job.probe.cancelled) jobs.remove (job);
			}
		}
		lock (cleaning) {
			cleaning = false;
		}
		if (jobs.size > 2) GLib.warning ("Simultaneous rendering: %d", jobs.size);
	}

	/**
	 * Init thread pool
	 */
	static void init_thread_pool () {
		lock (thread_pool) {
			if (thread_pool == null) {
				pushed_jobs = new LinkedList<RenderJob> ();
				try {
					thread_pool = new ThreadPool<RenderJob>.with_owned_data (static_render_image_surface, 4, false);
				}
				catch (ThreadError e) {
					stderr.printf ("%s\n", e.message);
					assert_not_reached ();
				}
			}
		}
	}

	/**
	 * Submit job to thread pool
	 */
	static void queue (RenderJob job) {
		try {
			lock (pushed_jobs) {pushed_jobs.add (job);}
			thread_pool.add (job);
		}
		catch (ThreadError e) {
			stderr.printf ("%s\n", e.message);
			assert_not_reached ();
		}
	}

	/**
	 * Used to show rendering progress
	 */
	int num_seed_rendered;

	/**
	 * Entry point for rendering process
	 */
	static void static_render_image_surface (owned RenderJob job) {
		job.render_image_surface ();
		lock (pushed_jobs) {pushed_jobs.remove (job);}
	}

	void render_image_surface () {
		if (probe.cancelled) return;
		var cr = new Cairo.Context (image_surface);
		cr.translate (origin_x, origin_y);
		for (var i = 0; i < border.size; i++) {
			num_seed_rendered = i;
			Pod motif_pod = border.motifs[i];
			var seed = border[i];
			image_surface_completed = false;
			Fractalizer.draw (motif_pod, cr, seed, probe, this);
			if (probe.cancelled) return;
		}
		image_surface_completed = true;
		render (1.0);
		// Set G_MESSAGES_DEBUG=all to get debug message printed
		if (Util.debug_on ()) GLib.print ("render time: %.2f\n", timer.elapsed ());
	}

	public void paint_image_surface (Cairo.Context cr, double x, double y) {
		cr.set_source_surface (image_surface, x, y);
		cr.paint ();
	}

	public void render (double fraction) {
		if (probe.cancelled) return;
		lock (pending_listeners) {
			if (pending_listeners == null) {
				pending_listeners = new ArrayList<IRender> ();
				messages = new ArrayList<double?> ();
			}
			lock (listeners) {
				if (listeners.size == 0) return;
				// If the list of pending is empty, that means there is no idle job yet.
				// So, we add one to service the listeners that we are about to add.
				if (pending_listeners.size == 0) {
					Idle.add (() => {
						render_pending_listeners ();
						return false;
					});
				}
				var message = (num_seed_rendered + fraction) / border.size;
				foreach (var listener in listeners) {
					if (pending_listeners.contains (listener)) {
						// Update the message
						messages[ pending_listeners.index_of (listener) ] = message;
					}
					else {
						pending_listeners.add (listener);
						messages.add (message);
					}
				}
			}//unlock
		}//unlock
	}

	/**
	 * This method is called during idle by the main thread
	 */
	public void render_pending_listeners () {
		lock (pending_listeners) {
			assert (pending_listeners.size == messages.size);
			for (var i = 0; i < pending_listeners.size; i++) {
				pending_listeners[i].render (messages[i]);
			}
			pending_listeners.clear ();
			messages.clear ();
		}
	}

	public void cancel (IRender? listener = null) {
		lock (listeners) {
			if (listener == null || !shared) {
				request_count --;
			}
			else if (shared && listener in listeners) {
				request_count --;
				listeners.remove (listener);
			}
			else {
				GLib.warning ("Miscancellation.");
			}
			if (request_count < 0) {
				GLib.warning ("Unexpected cancel.");
			}
			if (!shared) {
				probe.cancel ();
			}
		}
		remove_finished_jobs ();
	}
}
// vim: tabstop=4: textwidth=100: autoindent:
