/*  Mutiara motif designer.
	Copyright (C) 2011  Nor Jaidi Tuah

	This file is part of Mutiara.

	Mutiara is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
public abstract class Seed : Object {
	public bool debug = false;
	public bool is_affine = false;    // True for line of constant width
	static double epsilon = double.EPSILON;

	/**
	 * User configurable attributes
	 * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 */

	/**
	 * Lifespan of seed.
	 */
	public Role role = Role.DUAL;
	/**
	 * Where is the seed relative to the replications that will replace it?
	 * 0 = top
	 * 0.5 = middle
	 * 1 = bottom
	 */
	public double base_offset = 0.5;
	/**
	 * Symmetry type. Default is Flip.None & Flip.Y (used in Brunei traditional Air Muleh).
	 */
	public Flip flip_even = Flip.NONE;   // For segments 0, 2, 4, ...
	public Flip flip_odd = Flip.Y;       // For segments 1, 3, 5, ...
	public Tweak tweak = Tweak ();
	/**
	 * The width at t = 0.0
	 */
	public double width0 = 0.2;
	/**
	 * The width at t = 1.0
	 */
	public double width1 = 0.2;
	/**
	 * The curve is split. Each segment will be fractalized.
	 */
	public int num_segment = 4;
	/**
	 * 0 means non-fractal.
	 */
	public int level_of_detail = 10;
	/**
	 * Color is initially opaque black
	 */
	public RGBA color = RGBA (0.0, 0.0, 0.0, 1.0);
	/**
	 * 0 : don't mix.  1 : totally override
	 */
	public double color_mix = 0.5;
	public RenderStyle render_style = RenderStyle.THIN;

	/**
	 * Computed values used to speed up rendering
	 * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 *
	 * These values initialized by a call to start_render.
	 */

	/**
	 * Table used to generate "constant speed" curve rather than
	 * parametric curve.
	 */
	double[] distance_to_param;
	/**
	 * Number of intervals recorded in distance_to_param.
	 * Equal to distance_to_param.length - 1
	 */
	int div;
	/**
	 * Table used to indicate either a segment self-intersects.
	 * This is used (indirectly) in the calculation of too_thin.
	 */
	bool[] self_intersect;
	/**
	 * If any element in self_intersect is true, then
	 * twisted should be true, indicating that the seed requires
	 * special attention in the calculation of too_thin.
	 */
	bool twisted;
	/**
	 * Projections represented by each segment.
	 */
	public Projection[] projection;

	/**
	 * After creating a seed, call one of these to initialize.
	 */
	public void init_default (Point p) {
		init_default_sub (p);
	}

	public void init_full (double[] args) {
		init_full_sub (args);
	}

	protected abstract void init_default_sub (Point p);
	protected abstract void init_full_sub (double[] args);

	/**
	 * See Pod.serialize
	 */
	public string serialize (bool for_border = false) {
		if (for_border)
			return """%s
    segmentation %.1g, %d
    width %s, %s
    replication %d, %d, %d, %d%s
"""
				.printf (serialize_sub (),
					base_offset, num_segment,
					Util.format (width0, 4), Util.format (width1, 4),
					role, level_of_detail, flip_even, flip_odd, serialize_tweak ());

		else
			return """%s%s
    segmentation %.1g, %d
    width %s, %s
    replication %d, %d, %d, %d%s
    color (%.2g, %.2g, %.2g, %.2g) %.2g
    style %d
"""
				.printf (serialize_sub (),
					debug ? "\ndebug" : "",
					base_offset, num_segment,
					Util.format (width0, 4), Util.format (width1, 4),
					role, level_of_detail, flip_even, flip_odd, serialize_tweak (),
					color.r, color.g, color.b, color.a, color_mix,
					render_style);
	}

	string serialize_tweak () {
		if (tweak.equals (Tweak())) return "";
		else return "\n    tweak %s".printf (tweak.serialize ());
	}

	protected abstract string serialize_sub ();

	public void copy_parameters (Seed source) {
		base_offset = source.base_offset;
		num_segment = source.num_segment;
		width0 = source.width0;
		width1 = source.width1;
		role = source.role;
		level_of_detail = source.level_of_detail;
		flip_even = source.flip_even;
		flip_odd = source.flip_odd;
		tweak = source.tweak;
		color = source.color;
		color_mix = source.color_mix;
		render_style = source.render_style;
	}

	/**
	 * Compute (x(t), y(t)), where t is the distance (or time)
	 * of a "constant speed" curve.
	 */
	public Point point (double distance) {
		return point_raw (parameter (distance));
	}

	protected abstract Point point_raw (double t);

	/**
	 * Given a distance on a "constant speed" curve, convert it
	 * to a more convenient parameter for the computation of points on
	 * the curve.
	 */
	protected double parameter (double distance) {
		if (distance_to_param == null) return distance;
		if (distance < 0) {
			return distance * (distance_to_param[1] - distance_to_param[0]) * div;
		}
		if (distance >= 1) {
			return 1 + (distance - 1) * (distance_to_param[div] - distance_to_param[div - 1]) * div;
		}
		// index to distance_to_param table
		var n_exact = distance * div;
		var n = (int) n_exact;
		var t0 = distance_to_param[n];
		var t1 = distance_to_param[n + 1];
		var gap = t1 - t0;
		var t = t0 + (n_exact - n) * gap;
		if (t < 0) {
			GLib.warning ("unexpected value: t=%f", t);
		}
		return t;
	}

	/**
	 * To avoid unnecessary memory allocation, distance_to_param is populated
	 * only before rendering, and released after rendering.
	 */
	public void start_render () {
		// Don't assert (distance_to_param == null). Low-quality rendering recycle seeds.
		distance_to_param = null;
		init_distance_to_parametric ();
		init_self_intersect ();
		init_projection ();
	}

	public void stop_render () {
		distance_to_param = null;
		self_intersect = null;
		projection = null;
	}

	/**
	 * Populate distance_to_param table.
	 */
	protected virtual void init_distance_to_parametric () {
		// Compute absolute length
		var length = 0.0;
		var p0 = point_raw (0);
		var div_calc = 5000;      // Only for length calculation. Actual div is calculated below.
		bool normalized = true;   // Is this seed normalized to 0 .. 1 ?
		for (var i=1; i <= div_calc; i++) {
			var p1 = point_raw ((double) i / div_calc);
			if (Math.fabs (p1.x) > 3.0 && Math.fabs (p1.y) > 3.0) normalized = false;
			length += Point.hypot (p0, p1);
			p0 = p1;
		}
		if (length == 0) {
			distance_to_param = null;
			return;
		}

		// Now that we have the absolute length, we can compute the
		// distances normalized to 0..1
		// div divides the distance uniformly, unlike div_calc.
		div = int.max (100, (int)(normalized ? length * 1000 : length) / 5);
		distance_to_param = new double[div + 1];
		distance_to_param[0] = 0.0;
		distance_to_param[div] = 1.0;
		// Absolute measures:  0.0     i0,                          i0+1,len_1       div_calc,length
		// 0-1 measures:       0.0     dist_0        d,distance     dist_1           div,1.0
		var i0 = -1;
		var len_1 = 0.0;
		var dist_0 = 0.0;
		var dist_1 = 0.0;
		for (var d=1; d < div; d++) {
			var distance = (double) d / div;
			// Find the division number (i0) wherein point(distance) lies.
			while (dist_1 < distance) {
				i0 ++;
				assert (i0 + 1 <= div_calc);
				p0 = point_raw ((double) i0 / div_calc);
				var p1 = point_raw ((double) (i0 + 1) / div_calc);
				len_1 += Point.hypot (p0, p1);
				dist_0 = dist_1;
				dist_1 = len_1 / length;
			}
			if (!(dist_0 <= distance && distance <= dist_1)) {
				//GLib.warning ("%f %f %f", dist_0, distance, dist_1);
			}
			var t = (Math.fabs (dist_1 - dist_0) < epsilon) ? 
				(double) i0 / div_calc :
				(i0 + (distance - dist_0) / (dist_1 - dist_0)) / div_calc;
			if (t > 1) t = 1;
			if (t < 0) {
				GLib.warning ("unexpected value: t=%f (set to 0)", t);
				t = 0;
			}
			distance_to_param[d] = t;
		}
	}

	/**
	 * Check if segment n is self-intersecting
	 */
	public bool check_twisted (int n) {
		return self_intersect[n];
	}

	/**
	 * Initialize self_intersect table.
	 */
	void init_self_intersect () {
		self_intersect = new bool[num_segment];
		for (int i=0; i < num_segment; i++) {
			self_intersect[i] = check_self_intersect (i);
			twisted = (twisted || self_intersect[i]);
		}
	}

	/**
	 * Check if the normals striding segment n intersect each other
	 * or the curve joining their corresponding ends.
	 */
	bool check_self_intersect (int n) {
		var t0 = (double) n / num_segment;
		var t1 = (double) (n + 1) / num_segment;
		// Normal at t0
		var p0 = distant_point (t0, base_offset);
		var p1 = distant_point (t0, base_offset - 1);
		// Normal at t1
		var q0 = distant_point (t1, base_offset);
		var q1 = distant_point (t1, base_offset - 1);
		// Normals intersect each other?
		if (check_intersect (p0, p1, q0, q1)) return true;

		// Normals intersect top or bottom lines?
		// To check this, the top and bottom lines are segmented.
		// We perform 2 iterations in case the first iteration encounters
		// a line segment that exactly touches, but not crosses, a normal
		// (i.e., area_sign returns 0). The values of num_points in the
		// two iterations are co-primes to avoid the same touch.
		var num_points = int.max (50, div / num_segment);
		for (int k=0; k < 2; k++) {
			num_points++;  // Ensure num_points(0) and num_points(1) are co-primes.
			for (int i=0; i < num_points; i++) {
				var tA = t0 + (double) i / num_points / num_segment;
				var tB = t0 + (double) (i + 1) / num_points / num_segment;
				// Top and bottom segments
				var pA = distant_point (tA, base_offset);
				var pB = distant_point (tB, base_offset);
				var qA = distant_point (tA, base_offset - 1);
				var qB = distant_point (tB, base_offset - 1);
				if (check_intersect (p0, p1, pA, pB) ||
					check_intersect (p0, p1, qA, qB) ||
					check_intersect (q0, q1, pA, pB) ||
					check_intersect (q0, q1, qA, qB))
				{
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Is p1, p2, p3 clockwise (-ve area), anti-clockwise (+ve area), or colinear (zero) ?
	 * We want the sign rather than the actual area.
	 */
	static double area_sign (Point p1, Point p2, Point p3) {
		var area = (p2.x - p1.x) * (p3.y - p1.y) - (p3.x - p1.x) * (p2.y - p1.y);
		if (Math.fabs (area) <= epsilon) return 0;
		return area;
	}

	/**
	 * Are the two lines p1-p2 and q1-q2 intersecting?
	 */
	static bool check_intersect (Point p1, Point p2, Point q1, Point q2) {
		return
			// Are q1 and q2 on the opposite sides of p1-p2 ?
			area_sign (p1, p2, q1) * area_sign (p1, p2, q2) < 0 &&
			// Are p1 and p2 on the opposite sides of q1-q2 ?
			area_sign (q1, q2, p1) * area_sign (q1, q2, p2) < 0;
	}

	/**
	 * Precompute projections represented by each segment
	 */
	void init_projection () {
		projection = new Projection[num_segment];
		for (int i = num_segment - 1; i >= 0;  i--) {
			var odd = (i % 2 == 1);
			var t0 = (double) i / num_segment;
			var t1 = (double) (i + 1) / num_segment;
			projection[i] = new Projection.by_segment (this, t0, t1, odd ? flip_odd : flip_even);
		}
	}

	/**
	 * Compute the point x_distance along the seed and y_distance along the perpendicular
	 * at point (x_distance).
	 * y_distance is scaled according to width0 and width1.
	 */
	public abstract Point distant_point (double x_distance, double y_distance);

	/**
	 * Is the projected shape thinner than the given thickness?
	 * The thickness of the projected shape is calculated as area / perimeter.
	 * As a side effect, calculate "width" for rendering.
	 */
	public bool too_thin (Projection? pr, Probe? probe, double thickness, bool segment_twisted, ref double width) {
		var limit = 0.5 * thickness;

		// The figure is broken into 4 parts to cater for self intersection (twisted figure)
		//  a-----b
		//  |\   /|
		//  | \ / |
		//  |  x  |
		//  | / \ |
		//  |/   \|
		//  d-----c
		var area_ax = 0.0;
		var area_bx = 0.0;
		var area_cx = 0.0;
		var area_dx = 0.0;
		var area_ab = 0.0;
		var area_bc = 0.0;
		var area_cd = 0.0;
		var area_da = 0.0;
		var perimeter = 0.0;
		var dummy = 0.0;

		compute_area_perimeter (pr, probe, limit, 0, 0,  1, 0, ref area_ab, ref perimeter);
		compute_area_perimeter (pr, probe, limit, 1, 0,  1, 1, ref area_bc, ref perimeter);
		compute_area_perimeter (pr, probe, limit, 1, 1,  0, 1, ref area_cd, ref perimeter);
		compute_area_perimeter (pr, probe, limit, 0, 1,  0, 0, ref area_da, ref perimeter);
		var area = Math.fabs (area_ab + area_bc + area_cd + area_da);
		width = 3.5 * area / perimeter;

		// Figure too small (roughly equivalent to a circle with radius 1)?
		if (perimeter <= 6) return true;

		// Figure too thick even when self intersection is ignored
		if (width > thickness) return false;
		if (!twisted && !segment_twisted) return true;

		// Proceed below if the seed is twisted or about to be rendered onto a twisted segment
		compute_area_perimeter (pr, probe, limit, 0, 0,  0.5, 0.5, ref area_ax, ref dummy);
		compute_area_perimeter (pr, probe, limit, 1, 0,  0.5, 0.5, ref area_bx, ref dummy);
		compute_area_perimeter (pr, probe, limit, 1, 1,  0.5, 0.5, ref area_cx, ref dummy);
		compute_area_perimeter (pr, probe, limit, 0, 1,  0.5, 0.5, ref area_dx, ref dummy);

		area = Math.fabs (area_ab + area_bx - area_ax) +
		       Math.fabs (area_bc + area_cx - area_bx) +
		       Math.fabs (area_cd + area_dx - area_cx) +
		       Math.fabs (area_da + area_ax - area_dx);
		width = 3.5 * area / perimeter;
		return width <= thickness;
	}

	void compute_area_perimeter (Projection? pr, Probe? probe, double limit,
	                             double x0, double y0, double x1, double y1,
	                             ref double area, ref double perimeter)
	{
		y0 = base_offset - y0;
		y1 = base_offset - y1;
		var p0 = Projection.map (pr, distant_point (x0, y0));
		var p1 = Projection.map (pr, distant_point (x1, y1));
		compute_sub_area_perimeter (pr, probe, limit,
			x0, y0, p0,
			x1, y1, p1,
			Point.hypot (p0, p1),
			ref area, ref perimeter, 1000);
	}

	void compute_sub_area_perimeter (Projection? pr, Probe? probe, double limit,
	                             double x0, double y0, Point p0,
	                             double x1, double y1, Point p1,
	                             double distance,
	                             ref double area, ref double perimeter, int depth)
	{
		if (probe != null && probe.cancelled) return;
		while (true) {
			var xm = (x0 + x1) / 2;
			var ym = (y0 + y1) / 2;
			var pm = Projection.map (pr, distant_point (xm, ym));
			var distance_half = Point.hypot (p0, pm);
			if (distance <= limit && distance_half <= distance &&
				Point.hypot (p0, Projection.map (pr,
					distant_point (0.85 * x0 + 0.15 * x1, 0.85 * y0 + 0.15 * y1))) <= distance
				|| depth == 0)
			{
				area += p0.x * p1.y - p0.y * p1.x;
				perimeter += distance;
				return;
			}
			compute_sub_area_perimeter (pr, probe, limit, x0, y0, p0, xm, ym, pm,
			                            distance_half, ref area, ref perimeter, depth - 1);
			// compute_sub_area_perimeter (pr, probe, limit, xm, ym, pm, x1, y1, p1,
			//                             distance_etc, ref area, ref perimeter, depth - 1);
			x0 = xm; y0 = ym; p0 = pm; depth--;
			distance = Point.hypot (p0, p1);
		}
	}

	/**
	 * Draw by mapping every point onto the projection.
	 */
	public void draw (DeferredCairo dc, Projection? pr, Probe? probe, double limit) {
		var draft = limit > 2.0;   // If limit is course, assume only rough rendering is required
		if (render_style == RenderStyle.THIN || draft && render_style == RenderStyle.NONE) {
			var p = Projection.map (pr, point_raw (0.0));
			dc.move_to (p);
			draw_segment (dc, pr, probe, 0, 0, 1, 0, limit);
			dc.stroke ();
		}
		else if (render_style == RenderStyle.FILLED || render_style == RenderStyle.CLOSED) {
			var p = Projection.map (pr, point_raw (0.0));
			dc.move_to (p);
			draw_segment (dc, pr, probe, 0, 0, 1, 0, limit);
			if (pr != null) {
				// Draw a projected line segment to join the two end-points
				var line = new Line ();
				var p0 = point_raw (0);
				var p1 = point_raw (1);
				line.init_full ({p0.x, p0.y, p1.x, p1.y});
				line.draw_segment (dc, pr, probe, 1, 0, 0, 0, limit);
			}
			if (!draft && render_style == RenderStyle.FILLED) dc.fill ();
			else dc.stroke ();
		}
		else if (render_style == RenderStyle.THICK || render_style == RenderStyle.HOLLOW) {
			var p = Projection.map (pr, distant_point (0.0, base_offset));
			dc.move_to (p);
			draw_segment (dc, pr, probe, 0, base_offset, 1, base_offset, limit);
			draw_segment (dc, pr, probe, 1, base_offset, 1, base_offset - 1, limit);
			draw_segment (dc, pr, probe, 1, base_offset - 1, 0, base_offset - 1, limit);
			draw_segment (dc, pr, probe, 0, base_offset - 1, 0, base_offset, limit);
			if (!draft && render_style == RenderStyle.THICK) dc.fill ();
			else dc.stroke ();
		}
	}

	/**
	 * Draw the segmentation of the seed, showing the projection boxes.
	 */
	public void draw_segmentation (Cairo.Context cr, Projection? pr) {
		start_render ();
		var dc = new DeferredCairo ();
		// draw y = base - 0
		var p = Projection.map (pr, distant_point (0, base_offset));
		dc.move_to (p);
		draw_segment (dc, pr, null, 0, base_offset, 1, base_offset);
		dc.stroke ();
		dc.render (cr);
		// draw y = base - 1
		p = Projection.map (pr, distant_point (0, base_offset - 1));
		dc.move_to (p);
		draw_segment (dc, pr, null, 0, base_offset - 1, 1, base_offset - 1);
		dc.stroke ();
		dc.render (cr);
		// draw x = segment points
		for (var i = 0; i <= num_segment; i++) {
			var x = (double) i / num_segment;
			p = Projection.map (pr, distant_point (x, base_offset));
			dc.move_to (p);
			draw_segment (dc, pr, null, x, base_offset, x, base_offset - 1);
			dc.stroke ();
			dc.render (cr);
		}
		stop_render ();
	}

	/**
	 * Draw a projected line segment.
	 */
	void draw_segment (DeferredCairo dc, Projection? pr, Probe? probe,
	                   double x0, double y0, double x1, double y1, double limit = 0.5) {
	    var p0 = Projection.map (pr, distant_point (x0, y0));
	    var p1 = Projection.map (pr, distant_point (x1, y1));
		draw_sub_segment (dc, pr, probe,
			x0, y0, p0,
			x1, y1, p1,
			Point.hypot (p0, p1),
			limit, 1000);
	}

	/**
	 * Draw by recursively splitting into two smaller subsegments.
	 * p0 and p1 are pre-computed projections of (x0, y0) and (x1, y1) respectively.
	 */
	void draw_sub_segment (DeferredCairo dc, Projection? pr, Probe? probe,
	                       double x0, double y0, Point p0, double x1, double y1, Point p1,
	                       double distance, double limit, int depth) {
		if (probe != null && probe.cancelled) return;
		while (true) {
			if (depth == 0) return;
			var xm = (x0 + x1) / 2;
			var ym = (y0 + y1) / 2;
			var pm = Projection.map (pr, distant_point (xm, ym));
			var distance_half = Point.hypot (p0, pm);
			// Check multiple distances to cater for self-intersection
			if (distance <= limit && distance_half <= distance &&
				Point.hypot (p0, Projection.map (pr,
					distant_point (0.85 * x0 + 0.15 * x1, 0.85 * y0 + 0.15 * y1))) <= distance)
			{
				dc.line_to (p1);
				return;
			}
			draw_sub_segment (dc, pr, probe, x0, y0, p0, xm, ym, pm, distance_half, limit, depth - 1);
			// draw_sub_segment (dc, pr, probe, xm, ym, pm, x1, y1, p1, distance_etc, limit, depth - 1);
			x0 = xm; y0 = ym; p0 = pm; depth--;
			distance = Point.hypot (p0, p1);
		}
	}

	/**
	 * Variables and methods for interactive editing.
	 */
	double control_radius = 5;
	double control_magnetic_range = 10;
	double grid_magnetic_range = 10;
	protected int selected_control = -1;
	protected Point pressed_point;

	/**
	 * Draw directly using cairo methods.
	 */
	public abstract void draw_direct (Cairo.Context cr, double scale);

	/**
	 * Controls used for moving, scaling, rotation, etc.
	 */
	public abstract void draw_controls (Cairo.Context cr, double scale);

	protected void draw_square_control (Cairo.Context cr, Point p) {
		cr.rectangle (p.x - control_radius, p.y - control_radius, control_radius * 2, control_radius * 2);
		cr.fill ();
	}

	protected void draw_line_control (Cairo.Context cr, Point p0, Point p1) {
		cr.move_to (p0.x, p0.y);
		cr.line_to (p1.x, p1.y);
		cr.stroke ();
	}

	protected void draw_circle_control (Cairo.Context cr, Point p) {
		cr.arc (p.x, p.y, control_radius, 0, 2 * Math.PI);
		cr.fill ();
	}

	/**
	 * Is point 0 and point 1 within range?
	 * One of the point is a control, the other is the cursor.
	 */
	protected bool within_snap_range (Point p0, Point p1, double scale) {
		return Math.fabs (p0.x - p1.x) * scale <= control_magnetic_range &&
		       Math.fabs (p0.y - p1.y) * scale <= control_magnetic_range;
	}

	protected bool within_touch_range (Point p0, Point p1, double scale) {
		return Math.fabs (p0.x - p1.x) * scale <= control_radius &&
		       Math.fabs (p0.y - p1.y) * scale <= control_radius;
	}

	/**
	 * Is point p touching the seed
	 */
	protected bool touching (Point p, double scale) {
		// Divide into 4 segments to cater for very irregular shapes
		for (int i=0; i < 4; i++) {
			// Binary search for point on the seed closest to p
			var t0 = 0.25 * i;
			var t1 = 0.25 * (i + 1);
			while (t1 - t0 > epsilon) {
				var d0 = Point.hypot (p, point (t0)) * scale;
				var d1 = Point.hypot (p, point (t1)) * scale;
				if (d0 <= control_radius || d1 <= control_radius) {
					return true;
				}
				else if (d0 < d1) {
					t1 = (t0 + t1) / 2;
				}
				else {
					t0 = (t0 + t1) / 2;
				}
			}
		}
		return false;
	}

	/**
	 * See if any control is within the range of the cursor.
	 * Output: 
	 *  + selected_control
	 *            -1 : not selected
	 *           100 : drag the entire shape
	 *        others : determined by pressed_control
	 * + pressed_point : the point where the drag began
	 */
	public bool start_drag (Point p, double scale, bool selected, bool snap_on = false, double grid_size = 0) {
		selected_control = -1;
		save_drag_context ();
		var snapped_p = p;
		if (snap_on) {
			if (snap (ref snapped_p, scale, grid_size)) init_default (snapped_p);
		}
		if (selected) {
			selected_control = pressed_control (snapped_p, scale);
		}
		if (selected_control == -1 && touching (p, scale)) {
			selected_control = 100;
			pressed_point = p;
		}
		return selected_control != -1;
	}

	/**
	 * Which control is pressed and about to be dragged?
	 */
	protected abstract int pressed_control (Point p, double scale);
	protected abstract void save_drag_context ();

	public void drag (Point p, double scale, bool snap_on,
		              double grid_size, double snap_angle) requires (selected_control != -1) {
		if (selected_control == 100) {
			var delta_x = p.x - pressed_point.x;
			var delta_y = p.y - pressed_point.y;
			move_shape (delta_x, delta_y);
			if (snap_on) {
				double dx = 999999;  // TODO should be null
				double dy = 999999;  // TODO null
				distance_to_grid (scale, grid_size, out dx, out dy);
				if (dx != 999999) delta_x -= dx;
				if (dy != 999999) delta_y -= dy;
				move_shape (delta_x, delta_y);
			}
		}
		else {
			drag_control (p, scale, snap_on, grid_size, snap_angle);
		}
		// If the shape changes, recalculate its length
		if (selected_control != 100) init_distance_to_parametric ();
	}

	protected abstract void move_shape (double delta_x, double delta_y);
	protected abstract void drag_control (Point p, double scale, bool snap_on, double grid_size, double snap_angle);

	/**
	 * Snap to the nearest grid
	 */
	protected bool snap (ref Point p, double scale, double grid_size) {
		var bx = snap_x (ref p, scale, grid_size);
		var by = snap_y (ref p, scale, grid_size);
		return bx || by;
	}

	protected bool snap_x (ref Point p, double scale, double grid_size) {
		var grid_x = Math.round (p.x / grid_size) * grid_size;
		if (Math.fabs (p.x - grid_x) * scale <= grid_magnetic_range) {
			p.x = grid_x;
			return true;
		}
		else {
			return false;
		}
	}

	protected bool snap_y (ref Point p, double scale, double grid_size) {
		var grid_y = Math.round (p.y / grid_size) * grid_size;
		if (Math.fabs (p.y - grid_y) * scale <= grid_magnetic_range) {
			p.y = grid_y;
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Among the points returned by points_to_snap, find one that is closest to an x grid,
	 * and one that is closest to a y grid. Return the distances to the respective
	 * grid via the out parameters.
	 */
	protected void distance_to_grid (double scale, double grid_size, 
		                          out double distance_x, out double distance_y) {
		distance_x = 999999;  // TODO should be null
		distance_y = 999999;
		var ps = points_to_snap ();
		foreach (var p in ps) {
			var delta_x = p.x - Math.round (p.x / grid_size) * grid_size;
			if (Math.fabs (delta_x) * scale <= grid_magnetic_range) {
				if (distance_x == 999999 || Math.fabs (delta_x) < Math.fabs (distance_x)) {
					distance_x = delta_x;
				}
			}
			var delta_y = p.y - Math.round (p.y / grid_size) * grid_size;
			if (Math.fabs (delta_y) * scale <= grid_magnetic_range) {
				if (distance_y == 999999 || Math.fabs (delta_y) < Math.fabs (distance_y)) {
					distance_y = delta_y;
				}
			}
		}
	}

	/**
	 * When dragging the entire shape, which points should be snapped to the grids?
	 */
	protected abstract Point[] points_to_snap ();

	public abstract void draw_icon (Cairo.Context cr, double width, double height);
}
// vim: tabstop=4: textwidth=100: autoindent:
